import hashlib
import logging
import os
import threading
import time

from kubernetes import client, config

from ..constants import NAMESPACE, REPO2DOCKER_VERSION
from ..utils import (
    DOMAIN,
    DummyTask,
    new_user,
)
from .builder import ImageBuilderBase


def create_configmap(api_instance, configmap_name, data):
    body = {
        "apiVersion": "v1",
        "kind": "ConfigMap",
        "metadata": {"name": configmap_name},
        "data": data,
    }
    api_instance.create_namespaced_config_map(namespace=NAMESPACE, body=body)


def get_pod_logs(api_instance, pod_name, container_name, state):
    # wait for the pod to start
    timeout = 60 * 5
    start_time = time.time()
    while True:
        try:
            pod = api_instance.read_namespaced_pod(name=pod_name, namespace=NAMESPACE)
            if pod.status.phase in ("Running", "Succeeded"):
                break
            elif pod.status.phase == "Failed":
                print("Pod failed to start")
                return
            time.sleep(0.5)
            if timeout and time.time() - start_time > timeout:
                print("Timed out waiting for pod to start")
                return
        except client.exceptions.ApiException as e:
            if e.status != 404:
                print("Exception when calling CoreV1Api->read_namespaced_pod: %s\n" % e)
            pass

    try:
        pod_logs = api_instance.read_namespaced_pod_log(
            name=pod_name,
            namespace=NAMESPACE,
            container="r2d",
            follow=True,
            _preload_content=False,
        )
        for line in pod_logs:
            output = line.decode("utf-8").strip()
            if not (
                output.startswith("Using local repo")
                or output.startswith("[Repo2Docker]")
            ):
                state["state"].update(output.encode("utf-8"))
            if not state["dry_run"]:
                print(output, end="\n")
    except client.exceptions.ApiException as e:
        if e.status != 404:
            print("Exception when calling CoreV1Api->read_namespaced_pod_log: %s\n" % e)
        pass
    except Exception:
        pass


class KanikoImageBuilder(ImageBuilderBase):
    def __init__(self, gc, imageId=None, tale=None, auth=True):
        super().__init__(gc, imageId=imageId, tale=tale, auth=auth)

    def pull_r2d(self):
        pass

    def push_image(self, image):
        pass

    @property
    def engine(self):
        return (
            "--engine=kaniko "
            f"--KanikoEngine.registry_credentials=registry=registry.{DOMAIN} "
            "--KanikoEngine.registry_credentials=username=fido "
            "--KanikoEngine.registry_credentials=password=secretpass "
            f"--KanikoEngine.cache_registry=registry.{DOMAIN}/cache "
            "--KanikoEngine.cache_registry_credentials=username=fido "
            "--KanikoEngine.cache_registry_credentials=password=secretpass"
        )

    @staticmethod
    def _cleanup(pod_name, job_name, configmap_name, batch_api_instance, api_instance):
        batch_api_instance.delete_namespaced_job(name=job_name, namespace=NAMESPACE)
        api_instance.delete_namespaced_config_map(
            name=configmap_name, namespace=NAMESPACE
        )
        api_instance.delete_namespaced_pod(name=pod_name, namespace=NAMESPACE)

    def run_r2d(self, tag, dry_run=False, task=None):
        task = task or DummyTask
        suffix = new_user(8).lower()
        # Load Kubernetes configuration
        # config.load_kube_config()

        # Load in-cluster configuration
        config.load_incluster_config()

        # Define the directory to be uploaded
        local_directory_path = self.build_context
        configmap_name = f"job-configmap-{suffix}"

        # Read the contents of the local directory and create a ConfigMap
        # TODO: handle subdirectories
        data = {}
        for file_name in os.listdir(local_directory_path):
            file_path = os.path.join(local_directory_path, file_name)
            if os.path.isfile(file_path):
                with open(file_path, "r") as file:
                    data[file_name] = file.read()
            else:
                print(f"Skipping {file_path} as it is not a file.")

        # Create a ConfigMap with the contents of the local directory
        api_instance = client.CoreV1Api()
        create_configmap(api_instance, configmap_name, data)

        # Define Job manifest
        job_name = f"r2d-job-{suffix}"
        cmd = (
            f"mkdir -p {local_directory_path} && "
            f"cp -Lr /data/* {local_directory_path} && "
            f"{self.r2d_command(tag, dry_run=dry_run)}"
        )
        docker_secret_name = os.environ.get(
            "DOCKER_PULL_SECRET", "xarth-dockerhub-creds"
        )
        job_manifest = {
            "apiVersion": "batch/v1",
            "kind": "Job",
            "metadata": {"name": job_name},
            "spec": {
                "template": {
                    "metadata": {"name": "r2d"},
                    "spec": {
                        "containers": [
                            {
                                "name": "r2d",
                                "image": REPO2DOCKER_VERSION,
                                "command": ["/bin/bash", "-c", cmd],
                                "lifecycle": {
                                    "postStart": {
                                        "exec": {
                                            "command": [
                                                "/bin/sh",
                                                "-c",
                                                (
                                                    "echo $TRAEFIK_SERVICE_HOST registry.local.wholetale.org >> /etc/hosts && "
                                                    "mkdir -p /root/.docker && "
                                                    "cp /secrets/config.json /root/.docker/config.json"
                                                ),
                                            ]
                                        }
                                    }
                                },
                                "volumeMounts": [
                                    {
                                        "name": "job-volume",
                                        "mountPath": "/data",
                                    },
                                    {
                                        "name": "docker-secret",
                                        "mountPath": "/secrets/",
                                    },
                                ],
                                "workingDir": local_directory_path,  # Set working directory
                            }
                        ],
                        "imagePullSecrets": [{"name": docker_secret_name}],
                        "restartPolicy": "Never",
                        "volumes": [
                            {
                                "name": "job-volume",
                                "configMap": {"name": configmap_name},
                            },
                            {
                                "name": "docker-secret",
                                "secret": {
                                    "secretName": docker_secret_name,
                                    "items": [
                                        {
                                            "key": ".dockerconfigjson",
                                            "path": "config.json",
                                        }
                                    ],
                                },
                            },
                        ],
                    },
                },
                "backoffLimit": 1,
            },
        }

        # Create Job
        batch_api_instance = client.BatchV1Api()
        logging.info(f"Creating a r2d job in namespace '{NAMESPACE}'")
        batch_api_instance.create_namespaced_job(namespace=NAMESPACE, body=job_manifest)

        # Wait for the Job to complete
        logging.info("Waiting for the r2d job to complete... (5s)")
        time.sleep(5)
        logging.info("Listing associated pods")
        pods = api_instance.list_namespaced_pod(
            namespace=NAMESPACE, label_selector=f"job-name={job_name}"
        )
        pod_name = pods.items[0].metadata.name
        container_name = "r2d"
        logging.info(f"Pod name: {pod_name}")
        try:
            # Print Job logs while waiting
            logging.info("Starting logs thread (dry_run=%s)" % dry_run)
            state = {"state": hashlib.md5("R2D output".encode()), "dry_run": dry_run}
            logging.info("Printing logs (in a thread)")
            pod_thread = threading.Thread(
                target=get_pod_logs,
                args=(api_instance, pod_name, container_name, state),
            )
            pod_thread.start()

            while True:
                if task.canceled:
                    logging.info("Job canceled by user")
                    ret = {"StatusCode": -123, "error": "Canceled by user"}
                    self._cleanup(
                        pod_name,
                        job_name,
                        configmap_name,
                        batch_api_instance,
                        api_instance,
                    )
                    logging.info("Cleanup complete")
                    break

                job_status = batch_api_instance.read_namespaced_job_status(
                    name=job_name, namespace=NAMESPACE
                )
                if (
                    job_status.status.succeeded is not None
                    and job_status.status.succeeded > 0
                ):
                    logging.info("Job succeeded")
                    self._cleanup(
                        pod_name,
                        job_name,
                        configmap_name,
                        batch_api_instance,
                        api_instance,
                    )
                    logging.info("Cleanup complete")
                    ret = {"StatusCode": 0}
                    break
                elif (
                    job_status.status.failed is not None
                    and job_status.status.failed > 0
                ):
                    ret = {"StatusCode": -1, "error": "Job failed"}
                    break
                time.sleep(5)
        finally:
            # Stop the pod logs thread when done
            logging.info("Stopping logs thread")
            pod_thread.join()
            logging.info("Logs thread stopped")

        return ret, state["state"].hexdigest()



# Description: Context manager for changing the current working directory

import logging
import time
import docker
import requests
from requests.adapters import HTTPAdapter, Retry

from .constants import GIRDERFS_IMAGE, VOLUMES_ROOT
from .utils import stop_container


retries = Retry(
    connect=5,
    backoff_factor=0.5,
)


class FSContainer(object):
    @staticmethod
    def start_container(name):
        cli = docker.from_env()
        # Create container for handling FUSE mounts
        print("Creating WT Filesystem container...")
        fscontainer = cli.containers.run(
            image=GIRDERFS_IMAGE,
            name=name,
            detach=True,
            labels={"traefik.enable": "false"},
            mounts=[
                docker.types.Mount(
                    target=VOLUMES_ROOT,
                    source=VOLUMES_ROOT,
                    type="bind",
                    propagation="rshared",
                ),
            ],
            environment={
                "WT_VOLUMES_PATH": VOLUMES_ROOT,
            },
            devices=["/dev/fuse"],
            cap_add=["SYS_ADMIN"],
            security_opt=["apparmor:unconfined"],
            network="wt_celery",
            remove=True,
        )
        # wait for the container to be up and running
        # fail after 30s
        t = 0
        while True:
            time.sleep(1)
            try:
                fscontainer.reload()
            except docker.errors.NotFound:
                raise Exception("Failed to create WT Filesystem container")
            if fscontainer.status == "running":
                break
            if t > 30 or fscontainer.status == "exited":
                raise Exception("Failed to create WT Filesystem container")
            t += 1
        return fscontainer

    @staticmethod
    def mount(container, payload):
        # send payload to fscontainer using requests
        print("Sending payload to WT Filesystem container...")
        with requests.Session() as session:
            session.mount("http://", HTTPAdapter(max_retries=retries))
            response = session.post(
                f"http://{container.name}:8888/",
                json=payload,
                headers={"Content-Type": "application/json"},
            )
            response.raise_for_status()

    @staticmethod
    def stop_container(name):
        print("Sending shutdown request to WT Filesystem container...")
        cli = docker.from_env()
        try:
            container = cli.containers.get(name)
        except docker.errors.NotFound:
            return
        resp = requests.delete(f"http://{container.name}:8888/")
        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError:
            # log error from resp with traceback and continue
            logging.warning(
                f"Failed to shutdown WT Filesystem container {name} "
                f"with status code {resp.status_code}"
            )
            logging.warning(resp.text)
            pass

        stop_container(container)



"""WholeTale Girder Worker Plugin."""
import os
import docker
from girder_worker import GirderWorkerPluginABC
from kombu.common import Broadcast, Exchange, Queue


class GWVolumeManagerPlugin(GirderWorkerPluginABC):
    """Custom WT Manager providing WT tasks."""

    def __init__(self, app, *args, **kwargs):
        """Constructor."""
        self.app = app
        # Here we can also change application settings. E.g.
        # changing the task time limit:
        #
        queues = []
        if node_id := os.environ.get("SWARM_NODE_ID"):
            cli = docker.from_env()
            node = cli.nodes.get(node_id)
            if node.attrs["Spec"]["Role"] == "manager":
                queues.append(
                    Queue("manager", Exchange("manager", type="direct"), routing_key="manager")
                )

            if len(cli.nodes.list()) == 1 or node.attrs["Spec"]["Role"] != "manager":
                queues.append(
                    Queue("celery", Exchange("celery", type="direct"), routing_key="celery")
                )

            queues += [
                Queue(node_id, Exchange(node_id, type="direct"), routing_key=node_id),
            ]
        else:
            queues = [
                Queue(
                    "celery",
                    Exchange("celery", type="direct"),
                    routing_key="celery",
                ),
            ]
        queues.append(Broadcast("broadcast_tasks"))

        self.app.conf.task_queues = queues
        self.app.conf.task_routes = {
            "gwvolman.tasks.shutdown_container": {"queue": "broadcast_tasks"}
        }
        # self.app.config.update({
        #     'TASK_TIME_LIMIT': 300
        # })

    def task_imports(self):
        """Return a list of python importable paths."""
        return ["gwvolman.tasks"]



import base64
import hashlib
import json
import logging
import os
import shutil
import tempfile
from urllib.parse import urlparse

import requests
from packaging import version

from ..constants import R2D_FILENAMES
from ..utils import (
    DEPLOYMENT,
    _get_container_config,
    _get_stata_license_path,
)


class ImageBuilderBase:
    _build_context = None

    @property
    def build_context(self):
        if not self._build_context:
            self._build_context = self._create_build_context()
        return self._build_context

    @property
    def engine(self):
        # See https://github.com/whole-tale/repo2docker_wholetale/pull/44
        tag = self.container_config.repo2docker_version.rsplit(":")[-1]
        try:
            if version.parse(tag[1:]) < version.Version("1.2dev0"):
                return ""
        except version.InvalidVersion:
            # i.e. not something following v{version} which in our case
            # will be either "latest" or some specific manual tag
            pass
        return "--engine dockercli"

    def __init__(self, gc, imageId=None, tale=None, auth=True):
        if (imageId is None) == (tale is None):
            raise ValueError("Only one of 'imageId' and 'tale' can be set")

        self.gc = gc
        if tale is None:
            tale = {
                "_id": None,
                "imageId": imageId,
                "workspaceId": None,
                "config": {"extra_build_files": []},
            }
        self.tale = tale
        self.container_config = _get_container_config(gc, self.tale)

    def pull_r2d(self):
        raise NotImplementedError()

    def _create_build_context(self):
        temp_dir = tempfile.mkdtemp()
        logging.info(
            "Downloading r2d files to %s (taleId:%s)", temp_dir, self.tale["_id"]
        )
        extra_build_files = self.tale["config"].get("extra_build_files", [])
        workspaceId = self.tale.get("workspaceId")
        if workspaceId:
            if "**" in extra_build_files:
                # A special case when we want to have an entire workspace in the build context
                self.gc.downloadFolderRecursive(workspaceId, temp_dir)
            else:
                # Download standard r2d files
                for name in R2D_FILENAMES:
                    if item := next(self.gc.listItem(workspaceId, name=name), None):
                        self.gc.downloadItem(item["_id"], temp_dir)
                # Download any extra files specified by the Tale's config
                root_path = self.gc.get(
                    f"/resource/{workspaceId}/path", parameters={"type": "folder"}
                )
                for path in extra_build_files:
                    if resource := self.gc.get(
                        "/resource/lookup",
                        parameters={
                            "path": os.path.join(root_path, path),
                            "test": False,
                        },
                    ):
                        if resource["_modelType"] == "item":
                            self.gc.downloadItem(resource["_id"], temp_dir)
                        elif resource["_modelType"] == "folder":
                            self.gc.downloadFolderRecursive(resource["_id"], temp_dir)

        # Write the environment.json to the r2d context directory
        with open(os.path.join(temp_dir, "environment.json"), "w") as fp:
            json.dump(
                {
                    "config": {
                        "buildpack": self.container_config.buildpack,
                        "environment": self.container_config.environment,
                        "user": self.container_config.container_user,
                    }
                },
                fp,
            )
        return temp_dir

    def get_tag(self, force=False):
        """Compute a unique docker image tag.

        Tag is created as combination of 1) checksum of repo2docker files (apt.txt, etc)
        and the tale/image environment file, 2) checksum of Dockerfile created by r2d
        using files from 1).
        """
        env_hash = hashlib.md5("Environment checksum".encode())
        for root, dirs, files in os.walk(self.build_context):
            dirs.sort()
            for fname in sorted(files):
                env_hash.update(fname.encode())
                with open(os.path.join(root, fname), "rb") as fp:
                    env_hash.update(fp.read())
        if force and self.tale["_id"]:
            env_hash.update(self.tale["_id"].encode())

        # Perform dry run to get the Dockerfile's checksum
        registry_netloc = urlparse(DEPLOYMENT.registry_url).netloc
        logging.info("Computing tag for %s", self.tale["_id"])
        ret, output_digest = self.run_r2d(
            f"{registry_netloc}/placeholder_env/placeholder_dockerfile",
            dry_run=True,
        )
        logging.info("Computed tag %s", output_digest)
        if ret["StatusCode"] != 0:
            logging.error("Failed to compute a tag %s", ret)
            raise ValueError(f"Failed to compute a tag {ret=}")

        # Remove the temporary directory, cause we want entire workspace for build
        # NOTE: or maybe not? That would avoid bloating image with things we override anyway
        # shutil.rmtree(self.build_context, ignore_errors=True)

        return f"{registry_netloc}/tale/{env_hash.hexdigest()}:{output_digest}"

    def run_r2d(self, tag, dry_run=False, task=None):
        raise NotImplementedError()

    def push_image(self, image):
        raise NotImplementedError()

    def extra_args(self, dry_run=False):
        # Extra arguments for r2d
        extra_args = "--debug"
        if self.container_config.buildpack == "MatlabBuildPack":
            extra_args += " --build-arg FILE_INSTALLATION_KEY={}".format(
                os.environ.get("MATLAB_FILE_INSTALLATION_KEY")
            )
        elif self.container_config.buildpack == "StataBuildPack" and not dry_run:
            # License is also needed at build time but can't easily
            # be mounted. Pass it as a build arg
            with open(_get_stata_license_path(), "r") as license_file:
                stata_license = license_file.read()
                encoded = base64.b64encode(stata_license.encode("ascii")).decode(
                    "ascii"
                )
                extra_args += " --build-arg STATA_LICENSE_ENCODED='{}'".format(encoded)
        if base_image := os.environ.get("R2D_BASE_IMAGE"):
            extra_args += f" --Repo2Docker.base_image={base_image}"
        return extra_args

    def r2d_command(self, tag, dry_run=False):
        extra_args = self.extra_args(dry_run=dry_run)
        op = "--no-build" if dry_run else "--no-run"
        target_repo_dir = os.path.join(self.container_config.target_mount, "workspace")
        return (
            f"jupyter-repo2docker {self.engine} "
            "--config=/wholetale/repo2docker_config.py "
            f"--target-repo-dir={target_repo_dir} "
            f"--user-id=1000 --user-name={self.container_config.container_user} "
            f"--no-clean {op} {extra_args} "
            f"--image-name={tag} {self.build_context}"
        )

    def __del__(self):
        if self._build_context is not None:
            shutil.rmtree(self._build_context, ignore_errors=True)

    def cached_image(self, image):
        """Check if image exists in the registry"""
        _, full_name = image.split("/", 1)
        name, tag = full_name.split(":", 1)
        try:
            with requests.Session() as session:
                session.auth = (
                    os.environ.get("REGISTRY_USER", "fido"),
                    os.environ.get("REGISTRY_PASS"),
                )
                base_url = (
                    urlparse(DEPLOYMENT.registry_url)._replace(path="/v2/").geturl()
                )

                req = session.get(base_url)
                req.raise_for_status()

                req = session.get(
                    f"{base_url}{name}/manifests/{tag}",
                    headers={
                        "Accept": (
                            "application/vnd.docker.distribution.manifest.v2+json,"
                            "application/vnd.oci.image.manifest.v1+json"
                        )
                    },
                )
                req.raise_for_status()
                manifest = req.json()
                content_digest = req.headers["Docker-Content-Digest"]

                config_digest = manifest["config"]["digest"]

                req = session.get(
                    f"{base_url}{name}/blobs/{config_digest}",
                    headers={"Accept": manifest["config"]["mediaType"]},
                )
                req.raise_for_status()
                config = req.json()

                return {
                    "name": f"{urlparse(base_url).netloc}/{name}",
                    "tag": tag,
                    "digest": content_digest,
                    "created": config["created"],
                    "labels": config["config"]["Labels"],
                    "architecture": config["architecture"],
                    "os": config["os"],
                }
        except requests.exceptions.HTTPError as err:
            if err.response.status_code == 404:
                logging.info("Image %s not found in the registry", image)
                return
            raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import socket

API_VERSION = "2.1"
DEFAULT_USER = 1000
DEFAULT_GROUP = 100
ENABLE_WORKSPACES = True
MOUNTPOINTS = ["data"]
if ENABLE_WORKSPACES:
    # MOUNTPOINTS.append("workspace")
    MOUNTPOINTS.append("versions")
    MOUNTPOINTS.append("runs")


try:
    DEFAULT_GIRDER_API_URL = "http://" + socket.gethostbyname("girder") + ":8080/api/v1"
except socket.gaierror:
    DEFAULT_GIRDER_API_URL = "https://girder.dev.wholetale.org/api/v1"
GIRDER_API_URL = os.environ.get("GIRDER_API_URL", DEFAULT_GIRDER_API_URL)
LICENSE_PATH = os.environ.get("WT_LICENSE_PATH", "/licenses/")

REPO2DOCKER_VERSION = os.environ.get(
    "REPO2DOCKER_VERSION", "wholetale/repo2docker_wholetale:latest"
)
GIRDERFS_IMAGE = os.environ.get("GIRDERFS_IMAGE", "wholetale/girderfs:latest")
CPR_VERSION = os.environ.get("CPR_VERSION", "wholetale/wt-cpr:latest")
VOLUMES_ROOT = os.environ.get("WT_VOLUMES_PATH", "/mnt/homes")
NFS_PATH = os.environ.get("NFS_PATH", "/srv/vc_crypt/IMQCAM/girder")
NFS_SERVER = os.environ.get("NFS_SERVER")
NAMESPACE = os.environ.get("NAMESPACE", "wt")

RUN_WT_BUTTON_IMG = (
    "https://img.shields.io/badge/WholeTale-Run!-579ACA.svg?"
    "logo=data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABQAAAAUCAYAAACNiR0NAAAABHNCSVQICAgIfAhki"
    "AAAAAlwSFlzAAABDgAAAQ4B6Vk72QAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAAOGSURBVDiNj"
    "ZVdTJtVGMd/py+0IC0DNuhcwQKTC8OHY6WTZHGOxJiYmKGYLSQm6Mxu5jREix+7WGJ0xs+LSsbijdlQExSEscwroyNhY"
    "7ghxi1AFDdaRusG0zG2VunH+z7elDpsS3wuz3me3/n/n5zzHMUaISImoBLYmFi6BswopSRTTVYGUBnw+o2wb0/w9mRxK"
    "HITAulCEd+zYKI9ALvK6UC/61VaWCv+BbH3z3r7zbPhy6nVWG33s/DFc9Fygu2vqGU8mYE6qIfPevr3n8h0Ee2yYKzsJ"
    "51OeUYohHXbzCzOEY4upgsfahsN9vL245qSjuQYllEXh72Hdt/IdDHAyWPcTW0i09+1gneiQJgM2s8WtFCk3Oci793Ez"
    "einJ/rBdQLIjKtlPo4qVBEHL7F8Sv9E4csWx1tdP3kZvrP5bR2rWaNt3f8xcT1t4gbUUCxu/adiLNgS5VSas6UyDt4xn"
    "/cUrm+kU8vblsFe8Rpo6OxhI7GEnbcZyMU1Tk0fA8P3rt3xRtnfMcsQAeASURMC+GZPQuhKyjTk1xa+DsJ69i2gWuXzr"
    "HvcBf7DncxP3EOj3sDoajO6dl6rOb1AFwP/cYfYX+riCgTUBFcmirOtxQzfLUwCdvptHHy29MM/DBFXDeI6wb9o1N88/"
    "0QO502vvOHqShyJ/ODt6dKgHITsEmXGIW5pcwuRZMJDRtzOfXjryk9HDz/C9sduYSiBtmaHQCT0ogZEQCHCZCbk3k444"
    "+Tk/XvLdJMKaxkxBPvRBEnNzufLTl7WZyyAogSkc09PT2XOzs7eaL5KYo216DlWsmLh7m1dIuYxcZnY3MAtLnLyI7coW"
    "BdActZedjis8xMz3HyxCna29tpbW2tzAJm6uvr52OxmP3E171Ab4qiZ559HhHhi4/ezKja5XItAP4spZSISF91dfWLk5"
    "OTaZO/+vx4Zv9AbW0tVVVVXyqlZKVT73k8nmWz2Zy2wDAMDMNIu2c2m/F4PBHgQwATgFIqWFdXd9Dr9WK329dUc3fY7X"
    "a8Xi81NTWvrkyeVcNBRI7oun5gaGiIgYEBxsbGEFk9+pRSuN1uWlpaaGpqQtO0I0qplzKeKiLtIhIWERkZGZGGhgZxuV"
    "zicrnE7XbL6OioJCIsIimglHmYgDqA14CnBwcHHYFAAKUUpaWlNDc3B4F+4AOlVPB/Ae8CK6AC2JRYCgL+tb6AfwC+yH"
    "oMI9RIjwAAAABJRU5ErkJggg=="
)


# NOTE: changing order of ENV_FILES will result in cache invalidation.
# You have been warned.
R2D_FILENAMES = (
    "environment.yml",
    "Pipfile",
    "Pipfile.lock",
    "requirements.txt",
    "setup.py",
    "Project.toml",
    "REQUIRE",
    "install.R",
    "apt.txt",
    "DESCRIPTION",
    "postBuild",
    "start",
    "runtime.txt",
    "default.nix",
    "install.do",
    "JuliaProject.toml",
    "requirements3.txt",
    "toolboxes.txt",
    "Dockerfile",
)


class InstanceStatus(object):
    LAUNCHING = 0
    RUNNING = 1
    ERROR = 2


class TaleStatus(object):
    PREPARING = 0
    READY = 1
    ERROR = 2


class RunStatus(object):
    UNKNOWN = 0
    STARTING = 1
    RUNNING = 2
    COMPLETED = 3
    FAILED = 4
    CANCELED = 5


CREATE_VOLUME_STEP_TOTAL = 2
LAUNCH_CONTAINER_STEP_TOTAL = 2
UPDATE_CONTAINER_STEP_TOTAL = 2
IMPORT_TALE_STEP_TOTAL = 2
RECORDED_RUN_STEP_TOTAL = 4
BUILD_TALE_IMAGE_STEP_TOTAL = 2



# Copyright (c) 2016, Data Exploration Lab
# Distributed under the terms of the Modified BSD License.

"""A set of helper routines for WT related tasks."""

from collections import namedtuple
import os
import queue
import random
import re
import requests
import string
import time
import tempfile
import threading
import uuid
import logging
import docker
import datetime
import dateutil.relativedelta as rel

from .constants import (
    LICENSE_PATH,
    MOUNTPOINTS,
    NAMESPACE,
    REPO2DOCKER_VERSION,
    VOLUMES_ROOT,
    GIRDERFS_IMAGE,
)
from .lib.stats_collector import DockerStatsCollectorThread

DOCKER_URL = os.environ.get("DOCKER_URL", "unix://var/run/docker.sock")
MAX_FILE_SIZE = os.environ.get("MAX_FILE_SIZE", 200)
DOMAIN = os.environ.get("DOMAIN", "dev.wholetale.org")
TRAEFIK_ENTRYPOINT = os.environ.get("TRAEFIK_ENTRYPOINT", "websecure")
REGISTRY_USER = os.environ.get("REGISTRY_USER", "fido")
REGISTRY_PASS = os.environ.get("REGISTRY_PASS")
MOUNTS = {}
RETRIES = 5
container_name_pattern = re.compile(r"tmp\.([^.]+)\.(.+)\Z")
logger = logging.getLogger(__name__)

PooledContainer = namedtuple("PooledContainer", ["id", "path", "host"])
ContainerConfig = namedtuple(
    "ContainerConfig",
    [
        "buildpack",
        "repo2docker_version",
        "image",
        "command",
        "mem_limit",
        "cpu_shares",
        "container_port",
        "container_user",
        "target_mount",
        "url_path",
        "environment",
        "csp",
    ],
)

SIZE_NOTATION_RE = re.compile(r"^(\d+)([kmg]?b?)$", re.IGNORECASE)
SIZE_TABLE = {
    "": 1,
    "b": 1,
    "k": 1024,
    "kb": 1024,
    "m": 1024**2,
    "mb": 1024**2,
    "g": 1024**3,
    "gb": 1024**3,
}


def size_notation_to_bytes(size):
    if isinstance(size, int):
        return size
    match = SIZE_NOTATION_RE.match(size)
    if match:
        val, suffix = match.groups()
        return int(val) * SIZE_TABLE[suffix.lower()]
    raise ValueError


class K8SDeployment(object):
    """Container for WT-specific k8s stack deployment configuration."""

    __name__ = "K8SDeployment"
    dashboard_url = f"https://dashboard.{DOMAIN}"
    # girder_url = f"http://{os.environ.get('GIRDER_SERVICE_HOST')}:8080"
    girder_url = f"https://girder.{DOMAIN}"
    registry_url = f"https://registry.{DOMAIN}"
    builder_url = os.environ.get("BUILDER_URL", "https://builder.{DOMAIN}")
    existing_claim = os.environ.get("GIRDER_EXISTING_CLAIM", "girder-data")
    existing_claim_subpath = os.environ.get("GIRDER_EXISTING_CLAIM_SUBPATH", "")
    traefik_network = None
    tmpdir_mount = "/tmp"
    namespace = NAMESPACE
    girderfs_mount_type = os.environ.get("GIRDERFS_MOUNT_TYPE", "direct")
    ingress_class = os.environ.get("INGRESS_CLASS", "traefik")
    mounter_image = GIRDERFS_IMAGE


class DockerDeployment(object):
    """Container for WT-specific docker stack deployment configuration.

    This class allows to read and store configuration of services in a WT
    deployment. It's meant to be used as a singleton across gwvolman.
    """

    __name__ = "DockerDeployment"
    _dashboard_url = None
    _girder_url = None
    _registry_url = None
    builder_url = os.environ.get("BUILDER_URL", "https://builder.{DOMAIN}")
    _traefik_network = None
    _tmpdir_mount = None

    def __init__(self):
        self.docker_client = docker.from_env(version="1.28")

    @property
    def tmpdir_mount(self):
        """str: Path to the temporary directory used by gwvolman."""
        if self._tmpdir_mount is None:
            service = self.docker_client.services.get("wt_celery_worker")
            tmpdir = tempfile.gettempdir()
            mounts = service.attrs["Spec"]["TaskTemplate"]["ContainerSpec"]["Mounts"]
            self._tmpdir_mount = next(
                (_["Source"] for _ in mounts if _["Target"] == tmpdir), "/tmp"
            )
        return self._tmpdir_mount

    @property
    def traefik_network(self):
        """str: Name of the overlay network used by traefik for ingress."""
        if self._traefik_network is None:
            try:
                service = self.docker_client.services.get("wt_dashboard")
                self._traefik_network = service.attrs["Spec"]["Labels"][
                    "traefik.docker.network"
                ]
            except docker.errors.APIError:
                self._traefik_network = "wt_traefik-net"  # Default...
        return self._traefik_network

    @property
    def dashboard_url(self):
        """str: Dashboard's public url."""
        if self._dashboard_url is None:
            self._dashboard_url = self.get_host_from_traefik_rule("wt_dashboard")
        return self._dashboard_url

    @property
    def girder_url(self):
        """str: Girder's public url."""
        if self._girder_url is None:
            self._girder_url = self.get_host_from_traefik_rule("wt_girder")
        return self._girder_url

    @property
    def registry_url(self):
        """str: Docker Registry's public url."""
        if self._registry_url is None:
            self._registry_url = self.get_host_from_traefik_rule("wt_registry")
        return self._registry_url

    def get_host_from_traefik_rule(self, service_name):
        """Infer service's hostname from traefik frontend rule label.

        If services are unavailable (slave node), default to DOMAIN env settting
        """
        try:
            service = self.docker_client.services.get(service_name)
            ns = service.attrs["Spec"]["Labels"]["com.docker.stack.namespace"]
            router = service_name.replace("%s_" % ns, "")
            rule = service.attrs["Spec"]["Labels"][
                "traefik.http.routers.%s.rule" % router
            ]
            host = re.search(r"Host\(`(.+)`\)", rule).group(1)
            return "https://" + host
        except docker.errors.APIError:
            return "{}://{}.{}".format("https", service_name[3:], DOMAIN)


if os.environ.get("DEPLOYMENT", "docker") == "k8s":
    DEPLOYMENT = K8SDeployment()
else:
    DEPLOYMENT = DockerDeployment()
logger.warning(f"gwvolman:init: Using {DEPLOYMENT.__name__} as a Deployment backend")


def sample_with_replacement(a, size):
    """Get a random path."""
    return "".join([random.SystemRandom().choice(a) for x in range(size)])


def new_user(size):
    """Get a random path."""
    return sample_with_replacement(string.ascii_letters, size)


def _safe_mkdir(dest):
    try:
        os.mkdir(dest)
    except OSError as e:
        if e.errno != 17:
            raise
        logging.warn("Failed to mkdir {}".format(dest))
        pass


def _get_api_key(gc):
    api_key = None
    for key in gc.get("/api_key"):
        if key["name"] == "tmpnb" and key["active"]:
            api_key = key["key"]

    if api_key is None:
        api_key = gc.post("/api_key", data={"name": "tmpnb", "active": True})["key"]
    return api_key


def _get_user_and_instance(girder_client, instanceId):
    user = girder_client.get("/user/me")
    if user is None:
        logging.warn("Bad gider token")
        raise ValueError
    instance = girder_client.get("/instance/" + instanceId)
    return user, instance


def _get_container_config(gc, tale):
    if tale is None:
        container_config = {}  # settings['container_config']
    else:
        image = gc.get("/image/%s" % tale["imageId"])
        tale_config = image["config"] or {}
        if tale.get("config"):
            tale_config.update(tale["config"])

        image_info = tale.get("imageInfo", {})
        digest = image_info.get("digest")
        repo2docker_version = image_info.get("repo2docker_version", REPO2DOCKER_VERSION)

        try:
            mem_limit = size_notation_to_bytes(tale_config.get("memLimit", "2g"))
        except (ValueError, TypeError):
            mem_limit = 2 * 1024**3
        container_config = ContainerConfig(
            buildpack=tale_config.get("buildpack"),
            repo2docker_version=repo2docker_version,
            command=tale_config.get("command"),
            container_port=tale_config.get("port"),
            container_user=tale_config.get("user"),
            cpu_shares=tale_config.get("cpuShares"),
            environment=tale_config.get("environment"),
            image=digest,
            mem_limit=mem_limit,
            target_mount=tale_config.get("targetMount"),
            url_path=tale_config.get("urlPath"),
            csp=tale_config.get("csp"),
        )
    return container_config


def _launch_container(volume_info, container_config, gc):
    token = uuid.uuid4().hex
    # command
    if container_config.command:
        rendered_command = container_config.command.format(
            base_path="",
            port=container_config.container_port,
            ip="0.0.0.0",
            token=token,
        )
    else:
        rendered_command = None

    if container_config.url_path:
        rendered_url_path = container_config.url_path.format(token=token)
    else:
        rendered_url_path = ""

    logging.info("config = " + str(container_config))
    logging.info("command = " + str(rendered_command))
    cli = docker.from_env(version="1.28")
    cli.login(
        username=REGISTRY_USER, password=REGISTRY_PASS, registry=DEPLOYMENT.registry_url
    )
    # Fails with: 'starting container failed: error setting
    #              label on mount source ...: read-only file system'
    # mounts = [
    #     docker.types.Mount(type='volume', source=volumeName, no_copy=True,
    #                        target=container_config.target_mount)
    # ]

    # inject Girder token into the container
    environment = container_config.environment or []
    environment += [
        f"GIRDER_TOKEN={gc.token}",
        f"GIRDER_API_URL={gc.urlBase}",
        f"GIRDER_API_KEY={_get_api_key(gc)}",
    ]

    source_mount = os.path.join(VOLUMES_ROOT, "mountpoints", volume_info["volumeName"])
    mounts = []
    volumes = _get_container_volumes(source_mount, container_config, MOUNTPOINTS)
    user = gc.get("/user/me")
    volumes[os.path.join(VOLUMES_ROOT, f"homes/{user['login'][0]}/{user['login']}")] = {
        "bind": os.path.join(container_config.target_mount, "home"),
        "mode": "rw",
    }
    tale = gc.get("/tale/%s" % volume_info["taleId"])
    volumes[
        os.path.join(VOLUMES_ROOT, f"workspaces/{tale['_id'][0]}/{tale['_id']}")
    ] = {"bind": os.path.join(container_config.target_mount, "workspace"), "mode": "rw"}

    for source in volumes:
        mounts.append(
            docker.types.Mount(
                type="bind", source=source, target=volumes[source]["bind"]
            )
        )

    host = "tmp-{}".format(new_user(12).lower())
    environment.append(f"TMP_URL={host}.{DOMAIN}")

    # https://github.com/containous/traefik/issues/2582#issuecomment-354107053
    endpoint_spec = docker.types.EndpointSpec(mode="vip")

    # Use the specified CSP for iframes or default to deployed host
    csp = ""
    if container_config.csp:
        csp = container_config.csp
    else:
        csp = "frame-ancestors 'self' {}".format(DEPLOYMENT.dashboard_url)

    traefik_loadbalancer_prefix = f"traefik.http.services.{host}.loadbalancer"

    fqdn = f"{host}.{DOMAIN}"

    service = cli.services.create(
        container_config.image,
        command=rendered_command,
        labels={
            f"{traefik_loadbalancer_prefix}.server.port": str(
                container_config.container_port
            ),
            "traefik.enable": "true",
            "traefik.http.routers.%s.rule" % host: "Host(`{}.{}`)".format(host, DOMAIN),
            "traefik.http.routers.%s.entrypoints" % host: TRAEFIK_ENTRYPOINT,
            "traefik.http.routers.%s.tls" % host: "true",
            (
                f"traefik.http.middlewares.{host}"
                "-csp.headers.customresponseheaders.Content-Security-Policy"
            ): csp,
            f"{traefik_loadbalancer_prefix}.passhostheader": "true",
            "traefik.http.routers.%s.middlewares" % host: "girder, %s-csp" % host,
            "traefik.docker.network": DEPLOYMENT.traefik_network,
            "wholetale.instanceId": volume_info["instanceId"],
            "wholetale.taleId": volume_info["taleId"],
        },
        env=environment,
        mode=docker.types.ServiceMode("replicated", replicas=1),
        networks=[DEPLOYMENT.traefik_network],
        name=host,
        hosts={fqdn: "host-gateway"},
        mounts=mounts,
        endpoint_spec=endpoint_spec,
        constraints=["node.id == {}".format(volume_info["nodeId"])],
        resources=docker.types.Resources(mem_limit=container_config.mem_limit),
        restart_policy=docker.types.RestartPolicy(condition="none"),
    )

    # Wait for the server to launch within the container before adding it
    # to the pool or serving it to a user.
    # _wait_for_server(host_ip, host_port, path) # FIXME

    url = "{proto}://{host}.{domain}/{path}".format(
        proto="https", host=host, domain=DOMAIN, path=rendered_url_path
    )

    return service, {"url": url}


def _get_container_volumes(mountpoint, container_config, directories):
    volumes = {}
    for path in directories:
        source = os.path.join(mountpoint, path)
        target = os.path.join(container_config.target_mount, path)
        mode = "rw" if path in ("workspace", "home") else "ro"
        volumes[source] = {"bind": target, "mode": mode}

    if container_config.buildpack:
        # Mount the MATLAB and Stata runtime licenses
        if container_config.buildpack == "MatlabBuildPack":
            volumes[LICENSE_PATH] = {"bind": "/licenses"}
        elif container_config.buildpack == "StataBuildPack":
            # Weekly license expires each Sunday and is provided
            # in the format stata.YYYYMMDD.lic where YYYYMMDD is the
            # license expiration date.
            source_path = _get_stata_license_path()
            volumes[source_path] = {"bind": "/usr/local/stata/stata.lic"}
    return volumes


class DummyTask:
    canceled = False


def stop_container(container: docker.models.containers.Container):
    try:
        container.stop()
    except requests.exceptions.ReadTimeout:
        tries = 10
        while tries > 0:
            container.reload()
            if container.status == "exited":
                break
        if container.status != "exited":
            logging.error(f"Unable to stop container: {container.id}")
    except docker.errors.NotFound:
        logging.warning(f"Container {container.id} was already gone.")
    except docker.errors.DockerException as dex:
        logging.error(dex)
        raise


def _recorded_run(cli, mountpoint, container_config, tag, entrypoint, name, task=None):
    def logging_worker(log_queue, container):
        for line in container.logs(stream=True):
            log_queue.put(line.decode("utf-8").strip(), block=False)

    task = task or DummyTask
    log_queue = queue.Queue()
    print("Starting recorded run")

    # Configure container volumes for recorded run
    volumes = _get_container_volumes(
        mountpoint, container_config, ["data", "workspace"]
    )

    # TODO: use run config, not entrypoint
    run_cmd = f"sh {entrypoint}"

    print("Running Tale with command: " + run_cmd)
    print("Running image: " + tag)

    cli.images.pull(tag)

    container = cli.containers.create(
        image=tag,
        command=run_cmd,
        detach=True,
        name=name,
        volumes=volumes,
        working_dir=os.path.join(container_config.target_mount, "workspace"),
    )

    logging_thread = threading.Thread(
        target=logging_worker, args=(log_queue, container)
    )

    workspace_path = os.path.join(mountpoint, "workspace")
    dstats_tmppath = os.path.join(workspace_path, ".docker_stats")
    stats_thread = DockerStatsCollectorThread(container, dstats_tmppath)

    # Job output must come from stdout/stderr
    container.start()
    stats_thread.start()
    logging_thread.start()

    try:
        container = cli.containers.get(container.id)
        while container.status == "running":
            while not log_queue.empty():
                print(log_queue.get_nowait(), flush=True)
            if task.canceled:
                stop_container(container)
                break
            time.sleep(1)
            container = cli.containers.get(container.id)
    except docker.errors.NotFound:
        pass

    stats_thread.join()
    while not log_queue.empty():
        print(log_queue.get_nowait())
    logging_thread.join()

    if task.canceled:
        ret = {"StatusCode": -123}
    else:
        ret = container.wait()

    # Dump run std{out,err} and entrypoint used.
    with open(os.path.join(workspace_path, ".stdout"), "wb") as fp:
        fp.write(container.logs(stdout=True, stderr=False))
    with open(os.path.join(workspace_path, ".stderr"), "wb") as fp:
        fp.write(container.logs(stdout=False, stderr=True))
    with open(os.path.join(workspace_path, ".entrypoint"), "w") as fp:
        fp.write(entrypoint)
    try:
        container.remove()
    except docker.errors.NotFound:
        pass

    if not task.canceled and ret["StatusCode"] != 0:
        raise ValueError("Error executing recorded run")

    return ret


def _get_stata_license_path():
    license_date = datetime.date.today() + rel.relativedelta(days=1, weekday=rel.SU)
    return os.path.join(
        LICENSE_PATH, "stata", f"stata.{license_date.strftime('%Y%m%d')}.lic"
    )



from .tasks_docker import DockerTasks
from .tasks_kubernetes import KubernetesTasks

MAP = {"docker": DockerTasks, "k8s": KubernetesTasks}


class TasksFactory:
    def __init__(self, _type):
        self._check_type(_type)
        self._type = _type

    def _check_type(self, _type):
        if _type not in MAP:
            raise Exception("Unsupported deployment type: %s" % _type)

    def getTasksInstance(self):
        return MAP[self._type]()



import logging
import time
import uuid

import kubernetes
from kubernetes.client.rest import ApiException
from kubernetes.stream import stream

from .constants import (
    CREATE_VOLUME_STEP_TOTAL,
    LAUNCH_CONTAINER_STEP_TOTAL,
)
from .tasks_base import TasksBase
from .utils import (
    DOMAIN,
    K8SDeployment,
    _get_user_and_instance,
    new_user,
    _get_container_config,
    _get_api_key,
)
from .utils_k8s import tale_deployment, tale_service, tale_ingress


class KubernetesTasks(TasksBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        kubernetes.config.load_incluster_config()
        self.deployment = K8SDeployment()

    def create_volume(self, task, instance_id: str, mounts=None):
        user, instance = _get_user_and_instance(task.girder_client, instance_id)
        tale = task.girder_client.get("/tale/{taleId}".format(**instance))

        task.job_manager.updateProgress(
            message="Creating volume",
            total=CREATE_VOLUME_STEP_TOTAL,
            current=1,
            forceFlush=True,
        )

        vol_name = "_".join((tale["_id"], user["login"], new_user(6)))

        task.job_manager.updateProgress(
            message="Volume created",
            total=CREATE_VOLUME_STEP_TOTAL,
            current=CREATE_VOLUME_STEP_TOTAL,
            forceFlush=True,
        )

        return {
            "nodeId": None,
            "fscontainerId": None,
            "mounts": mounts,
            "volumeName": vol_name,
            "instanceId": instance["_id"],
            "taleId": tale["_id"],
        }

    def _ensure_one(self, list, type, instanceId):
        if len(list) == 0:
            # TODO: use exception
            # The girder side happily creates the instance even if launch_container throws
            # an exception, which it shouldn't.
            # raise LookupError('No %s found for instanceId %s' % (type, instanceId))
            logging.warning("No %s found for instanceId %s" % (type, instanceId))
            return None
        if len(list) > 1:
            raise LookupError(
                "Multiple %s found for instanceId %s" % (type, instanceId)
            )

        return list[0]

    def remove_volume(self, task, instanceId):
        return

    def _wait_for_pod(self, instanceId):
        api = kubernetes.client.CoreV1Api()
        tic = time.time()
        timeout = 30.0

        # wait until task is started
        while time.time() - tic < timeout:
            pods = api.list_namespaced_pod(
                self.deployment.namespace, label_selector="instanceId=%s" % instanceId
            )
            if len(pods.items) == 0:
                logging.info("_wait_for_pod with instance id %s no match" % instanceId)
            elif len(pods.items) > 1:
                logging.error(
                    "_wait_for_pod %s multiple matches; this should not be happening"
                    % instanceId
                )
            else:
                pod = pods.items[0]
                if pod.status.phase == "Running":
                    return pod
                elif pod.status.phase == "Pending":
                    # Reset deadline
                    tic = time.time()
                elif pod.status.phase == "Failed":
                    raise Exception(
                        "Pod %s failed. Reason: %s, message: %s"
                        % (instanceId, pod.status.reason, pod.status.message)
                    )
            time.sleep(5)
        raise Exception("Pod %s startup timed out" % instanceId)

    def _render_config(self, container_config):
        token = uuid.uuid4().hex
        # command
        if container_config.command:
            rendered_command = container_config.command.format(
                base_path="",
                port=container_config.container_port,
                ip="0.0.0.0",
                token=token,
            )
        else:
            rendered_command = None

        if container_config.url_path:
            rendered_url_path = container_config.url_path.format(token=token)
        else:
            rendered_url_path = ""
        return container_config._replace(
            command=rendered_command, url_path=rendered_url_path
        )

    def launch_container(self, task, payload):
        """Launch a container using a Tale object."""
        instanceId = payload["instanceId"]
        user, instance = _get_user_and_instance(task.girder_client, instanceId)
        tale = task.girder_client.get("/tale/{taleId}".format(**instance))

        task.job_manager.updateProgress(
            message="Starting container",
            total=LAUNCH_CONTAINER_STEP_TOTAL,
            current=1,
            forceFlush=True,
        )

        print("Launching container for a Tale...")
        if "imageInfo" not in tale:
            # Wait for image to be built
            tic = time.time()
            timeout = 180.0
            time_interval = 5

            while time.time() - tic < timeout:
                tale = task.girder_client.get("/tale/{taleId}".format(**instance))
                if "imageInfo" in tale and "digest" in tale["imageInfo"]:
                    break
                msg = f"Waiting for image build to complete. ({time_interval}s)"
                logging.info(msg)
                print(msg)
                time.sleep(5)

        host = f"tmp{new_user(12).lower()}"
        deployment_name = "tale-" + host

        # Must match ingress host
        service_name = host

        girder_api_url = f"{self.deployment.girder_url}/api/v1"
        template_params = {
            "deploymentName": deployment_name,
            "host": host,
            "domain": DOMAIN,
            "ingressClass": K8SDeployment.ingress_class,
            "deploymentNamespace": self.deployment.namespace,
            "claimName": K8SDeployment.existing_claim,
            "claimSubPath": K8SDeployment.existing_claim_subpath,
            "girderApiUrl": girder_api_url,
            "girderApiKey": _get_api_key(task.girder_client),
            "mounterImage": self.deployment.mounter_image,
            "instanceId": instanceId,
            "girderToken": task.girder_client.token,
            "girderFSMountType": K8SDeployment.girderfs_mount_type,
            "homeSubPath": f"homes/{user['login'][0]}/{user['login']}",
            "workspaceSubPath": f"workspaces/{tale['_id'][0]}/{tale['_id']}",
        }

        container_config = _get_container_config(task.girder_client, tale)

        container_config = self._render_config(container_config)
        self._render_config(container_config)
        template_params.update(
            {
                "command": container_config.command,
                "mountPoint": container_config.target_mount,
                "instancePort": container_config.container_port,
                "instanceImage": container_config.image,
            }
        )

        # girderfsDef
        template_params["girderfsDef"] = {
            "mounts": payload["mounts"],
            "girderApiUrl": girder_api_url,
            "girderApiKey": _get_api_key(task.girder_client),
            "girderToken": task.girder_client.token,
            "taleId": tale["_id"],
            "userId": user["_id"],
            "mountType": K8SDeployment.girderfs_mount_type,
            "root": "/",
        }

        # create deployment and service
        tale_deployment(template_params)

        # wait until task is started
        pod = self._wait_for_pod(instanceId)

        # exec girderfs-mount in mounter container
        self._execute_girderfs(pod, "girderfs-mount")

        tale_service(template_params)
        tale_ingress(template_params)

        print("Environment is up and running.")
        task.job_manager.updateProgress(
            message="Container started",
            total=LAUNCH_CONTAINER_STEP_TOTAL,
            current=LAUNCH_CONTAINER_STEP_TOTAL,
            forceFlush=True,
        )

        payload["url"] = f"https://{host}.{DOMAIN}/{container_config.url_path}"
        payload["name"] = service_name
        return payload

    def _execute_girderfs(self, pod, cmd=None):
        # exec into mounter container in the pod
        api = kubernetes.client.CoreV1Api()
        try:
            resp = stream(
                api.connect_get_namespaced_pod_exec,
                name=pod.metadata.name,
                namespace=self.deployment.namespace,
                command=cmd,
                container="mounter",
                stderr=True,
                stdin=False,
                stdout=True,
                tty=False,
            )
            print("Response: " + resp)
        except ApiException as e:
            print(
                "Exception when calling CoreV1Api->connect_get_namespaced_pod_exec: %s\n"
                % e
            )

    def shutdown_container(self, task, instanceId):
        """Shutdown a running Tale."""
        logging.info("Shutting down container for instance %s" % instanceId)
        api = kubernetes.client.AppsV1Api()
        deployments = api.list_namespaced_deployment(
            namespace=self.deployment.namespace,
            label_selector=f"instanceId={instanceId}",
        )
        deployment = self._ensure_one(deployments.items, "deployments", instanceId)
        if deployment is None:
            return

        # umount girderfs
        # pod = self._wait_for_pod(instanceId)
        # self._execute_girderfs(pod, "girderfs-umount")

        api.delete_namespaced_deployment(
            name=deployment.metadata.name, namespace=self.deployment.namespace
        )

        api = kubernetes.client.CoreV1Api()
        services = api.list_namespaced_service(
            namespace=self.deployment.namespace,
            label_selector="instanceId=%s" % instanceId,
        )
        service = self._ensure_one(services.items, "services", instanceId)
        if service is not None:
            api.delete_namespaced_service(
                name=service.metadata.name, namespace=self.deployment.namespace
            )

        api = kubernetes.client.NetworkingV1Api()
        ingresses = api.list_namespaced_ingress(
            namespace=self.deployment.namespace,
            label_selector="instanceId=%s" % instanceId,
        )
        ingress = self._ensure_one(ingresses.items, "ingresses", instanceId)
        if ingress is not None:
            api.delete_namespaced_ingress(
                name=ingress.metadata.name, namespace=self.deployment.namespace
            )



import logging
import os

from .docker import DockerImageBuilder  # noqa
from .kaniko import KanikoImageBuilder  # noqa
from .remote import RemoteImageBuilder  # noqa

logger = logging.getLogger(__name__)

if os.environ.get("BUILDER_URL"):
    ImageBuilder = RemoteImageBuilder
elif os.environ.get("DEPLOYMENT", "docker") == "k8s":
    ImageBuilder = KanikoImageBuilder
else:
    ImageBuilder = DockerImageBuilder

logger.warning(f"gwvolman:init: Using {ImageBuilder.__name__} as image builder")



"""A set of WT related Girder tasks."""
import logging
import os
import time

from girder_worker.app import app
from girder_worker.utils import girder_job

from .r2d import ImageBuilder
from .tasks_factory import TasksFactory

# from girder_worker.plugins.docker.executor import _pull_image
from .utils import (
    _safe_mkdir,
)

tasks = TasksFactory(os.environ.get("DEPLOYMENT", "docker")).getTasksInstance()


@girder_job(title="Create Tale Data Volume")
@app.task(bind=True)
def create_volume(task, instance_id, mounts):
    return tasks.create_volume(task, instance_id, mounts=mounts)


@girder_job(title="Spawn Instance")
@app.task(bind=True)
def launch_container(task, service_info):
    return tasks.launch_container(task, service_info)


@girder_job(title="Update Instance")
@app.task(bind=True)
def update_container(task, instanceId, digest=None):
    return tasks.update_container(task, instanceId, digest=digest)


@girder_job(title="Shutdown Instance")
@app.task(bind=True)
def shutdown_container(task, instanceId):
    return tasks.shutdown_container(task, instanceId)


@girder_job(title="Remove Tale Data Volume")
@app.task(bind=True)
def remove_volume(task, instanceId):
    return tasks.remove_volume(task, instanceId)


@girder_job(title="Build Tale Image")
@app.task(bind=True)
def build_tale_image(task, tale_id, force=False):
    return tasks.build_tale_image(task, tale_id, force=force)


@girder_job(title="Publish Tale")
@app.task(bind=True)
def publish(task, tale_id, token, version_id, repository=None, draft=False):
    return tasks.publish(
        task, tale_id, token, version_id, repository=repository, draft=draft
    )


@girder_job(title="Import Tale")
@app.task(bind=True)
def import_tale(task, lookup_kwargs, tale, spawn=True):
    return tasks.import_tale(task, lookup_kwargs, tale, spawn=spawn)


@app.task(bind=True)
def rebuild_image_cache(self):
    logging.info("Rebuilding image cache")

    # Get the list of images
    images = self.girder_client.get("/image")

    # Build each base image
    for image in images:
        image_builder = ImageBuilder(self.girder_client, imageId=image["_id"])
        tag = image_builder.get_tag()
        container_config = image_builder.container_config

        logging.info(
            "Building %s %s in %s with %s",
            image["name"],
            tag,
            image_builder.build_context,
            container_config.repo2docker_version,
        )

        start = time.time()
        ret, _ = image_builder.run_r2d(tag)

        elapsed = int(time.time() - start)
        if ret["StatusCode"] != 0:
            logging.error("Error building %s", image["name"])
        else:
            logging.info("Build time: %i seconds", elapsed)


def _make_fuse_dirs(mountpoint, directories):
    """Create fuse directories"""
    for suffix in directories:
        _safe_mkdir(os.path.join(mountpoint, suffix))


@girder_job(title="Recorded Run")
@app.task(bind=True)
def recorded_run(task, run_id, tale_id, entrypoint):
    return tasks.recorded_run(task, run_id, tale_id, entrypoint)


@app.task()
def check_on_run(run_state):
    return tasks.check_on_run(run_state)


@girder_job(title="Clean failed Recorded Run")
@app.task(bind=True)
def cleanup_run(task, run_id):
    return tasks.cleanup_run(task, run_id)



import json
import os

import requests

from .builder import ImageBuilderBase
from ..utils import DOMAIN


class RemoteImageBuilder(ImageBuilderBase):
    def __init__(
        self,
        gc,
        imageId=None,
        tale=None,
        builder_url=None,
        registry_user=None,
        registry_password=None,
        registry_url=None,
        auth=True,
    ):
        super().__init__(gc, imageId=imageId, tale=tale, auth=auth)
        self.builder_url = builder_url or os.environ.get(
            "BUILDER_URL", "https://builder.local.xarthisius.xyz"
        )
        self.registry_url = registry_url or f"https://registry.{DOMAIN}"
        self.registry_user = registry_user or os.environ.get("REGISTRY_USER", "fido")
        self.registry_password = registry_password or os.environ.get("REGISTRY_PASS")

    def pull_r2d(self):
        response = requests.put(
            f"{self.builder_url}/pull",
            params={
                "repo2docker_version": self.container_config.repo2docker_version,
            },
            stream=True,
        )
        for chunk in response.iter_lines():  # Adjust chunk size as needed
            try:
                msg = json.loads(chunk)
                try:
                    print(msg["status"])
                except KeyError:
                    raise json.jsonJSONDecodeError
            except json.JSONDecodeError:
                print(chunk)

    def push_image(self, image):
        """Push image to the registry"""
        repository, tag = image.split(":", 1)
        response = requests.put(
            f"{self.builder_url}/push",
            params={
                "image": image,
                "registry_user": self.registry_user,
                "registry_password": self.registry_password,
                "registry_url": f"https://registry.{DOMAIN}",
            },
            stream=True,
        )
        for chunk in response.iter_lines():  # Adjust chunk size as needed
            print(chunk)

    def run_r2d(self, tag, dry_run=False, task=None):
        """
        Run repo2docker on the workspace using a shared temp directory. Note that
        this uses the "local" provider.  Use the same default user-id and
        user-name as BinderHub
        """
        response = requests.post(
            f"{self.builder_url}/build",
            params={
                "taleId": self.tale["_id"],
                "apiUrl": f"https://girder.{DOMAIN}/api/v1",
                "token": self.gc.token,
                "registry_url": self.registry_url,
                "dry_run": dry_run,
                "tag": tag,
            },
            stream=True,
        )
        for chunk in response.iter_lines():
            try:
                msg = json.loads(chunk)
                if "message" in msg:
                    msg = msg["message"]
                    if isinstance(msg, dict) and "error" in msg.keys():
                        return {"StatusCode": 1, "error": msg["error"]}, None
                    print(msg)
                elif "return" in msg:
                    data = msg["return"]
                    return data["ret"], data["digest"]
                elif "error" in msg:
                    return {"StatusCode": 1, "error": msg["error"]}, None
            except json.JSONDecodeError:
                print(chunk)



import hashlib
import json
import logging
import os
import tempfile

import docker
from fastapi import FastAPI, HTTPException, Query
from fastapi.responses import StreamingResponse
from girder_client import GirderClient

from ..r2d.docker import DockerImageBuilder

app = FastAPI()
client = docker.from_env()
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


@app.put("/pull")
async def pull_docker_r2d_image(
    repo2docker_version: str = Query(
        ..., description="Repository and version of the image"
    ),
):
    async def pull_stream():
        try:
            for line in client.api.pull(repository=repo2docker_version, stream=True):
                line = json.loads(line.decode("utf-8").strip())
                yield json.dumps(line) + "\n"
        except Exception as e:
            yield json.dumps({"error": str(e)}) + "\n"

    return StreamingResponse(pull_stream(), media_type="application/json")


@app.put("/push")
async def push_tale_image(
    image: str = Query(..., description="Repository and version of the image"),
    registry_url: str = Query(..., description="Docker registry URL"),
    registry_user: str = Query(..., description="Docker registry username"),
    registry_password: str = Query(..., description="Docker registry password"),
):
    async def push_stream():
        try:
            repository, tag = image.split(":", 1)
            client.api.login(
                registry=registry_url,
                username=registry_user,
                password=registry_password,
            )
            for line in client.api.push(repository, tag=tag, stream=True, decode=True):
                yield json.dumps(line) + "\n"
        except Exception as e:
            yield json.dumps({"error": str(e)}) + "\n"

    return StreamingResponse(push_stream(), media_type="application/json")


@app.post("/build")
async def build_tale(
    taleId: str = Query(..., description="Tale identifier"),
    apiUrl: str = Query(..., description="Girder API URL"),
    token: str = Query(..., description="Girder authentication token"),
    registry_url: str = Query(..., description="Docker registry URL"),
    dry_run: bool = Query(..., description="If true, do not build the image"),
    tag: str = Query(..., description="Repository and version of the image"),
):
    girder_client = GirderClient(apiUrl=apiUrl)
    girder_client.token = token
    try:
        tale = girder_client.get("tale/%s" % taleId)
    except Exception:
        raise HTTPException(status_code=400, detail="Invalid Tale ID")

    image_builder = DockerImageBuilder(
        girder_client, tale=tale, registry_url=registry_url, auth=False
    )

    async def build_stream():
        try:
            yield json.dumps({"message": f"Building image {tag}"}) + "\n"
            r2d_cmd = image_builder.r2d_command(tag, dry_run=dry_run)
            r2d_context_dir = os.path.relpath(
                image_builder.build_context, tempfile.gettempdir()
            )
            host_r2d_context_dir = os.path.join("/tmp", r2d_context_dir)

            volumes = {
                host_r2d_context_dir: {
                    "bind": image_builder.build_context,
                    "mode": "rw",
                },
                "/var/run/docker.sock": {
                    "bind": "/var/run/docker.sock",
                    "mode": "rw",
                },
            }

            docker_group = os.environ.get("DOCKER_GROUP", "docker")
            if docker_group.isdigit():
                docker_group = int(docker_group)

            container = client.containers.run(
                image=image_builder.container_config.repo2docker_version,
                command=r2d_cmd,
                group_add=[docker_group],
                environment={"DOCKER_HOST": "unix:///var/run/docker.sock"},
                privileged=True,
                detach=True,
                remove=False,
                volumes=volumes,
            )

            yield json.dumps({"message": f"Calling {r2d_cmd}"}) + "\n"

            h = hashlib.md5("R2D ouptut".encode())
            for line in container.logs(stream=True):
                output = line.decode("utf-8").strip()
                if not output.startswith("Using local repo"):
                    h.update(output.encode("utf-8"))
                    logger.info(output)
                if not dry_run:
                    yield json.dumps({"message": output}) + "\n"

            try:
                ret = container.wait(timeout=10)
            except (docker.errors.TimeoutError, docker.errors.NotFound):
                ret = {"StatusCode": -123}

            if ret["StatusCode"] != 0:
                yield json.dumps({"error": f"Error building image {tag}"}) + "\n"

            yield json.dumps({"return": {"ret": ret, "digest": h.hexdigest()}}) + "\n"
        except Exception as e:
            yield json.dumps({"error": str(e)}) + "\n"

    return StreamingResponse(build_stream(), media_type="application/json")



import math
import time
from threading import Thread

import docker


class DockerStatsCollectorThread(Thread):
    def __init__(self, container, output_path):
        super(DockerStatsCollectorThread, self).__init__()
        self.daemon = True
        self.container = container
        self.output_path = output_path

    def container_finished(self, ts):
        try:
            if ts == "0001-01-01T00:00:00Z":
                self.container.reload()
            return self.container.attrs["State"]["Status"] not in ("created", "running")
        except docker.errors.NotFound:
            return True

    def run(self):
        while True:
            try:
                d = self.container.stats(stream=False)
            except docker.errors.NotFound:
                break
            ts = d["read"]

            if self.container_finished(ts):
                break

            if ts != "0001-01-01T00:00:00Z":
                mem_usage, mem_limit = self.calculate_memory(d)
                bytes_in, bytes_out = self.calculate_network_bytes(d)
                blkio_rd, blkio_wr = self.calculate_blkio_bytes(d)
                line = (
                    f"{ts} - {self.calculate_cpu_percent(d):.2f}%, {mem_usage} / {mem_limit},"
                    f" {bytes_in} / {bytes_out}, {blkio_rd} / {blkio_wr},"
                    f" {d.get('pids_stats', {}).get('current', 0)}\n"
                )
                with open(self.output_path, mode="a") as fp:
                    fp.write(line)
            time.sleep(5)

    @staticmethod
    def convert_size(size_bytes, binary=True):
        if size_bytes == 0:
            return "0B"
        if binary:
            suffix = "i"
            base = 1024
        else:
            suffix = ""
            base = 1000
        size_name = (
            "B",
            f"K{suffix}B",
            f"M{suffix}B",
            f"G{suffix}B",
            f"T{suffix}B",
            f"P{suffix}B",
        )
        i = int(math.floor(math.log(size_bytes, base)))
        p = math.pow(base, i)
        s = round(size_bytes / p, 2)
        return "%s %s" % (s, size_name[i])

    @staticmethod
    def calculate_cpu_percent(d):
        cpu_count = len(d["cpu_stats"]["cpu_usage"].get("percpu_usage", [1]))
        cpu_percent = 0.0
        cpu_delta = float(d["cpu_stats"]["cpu_usage"]["total_usage"]) - float(
            d["precpu_stats"]["cpu_usage"]["total_usage"]
        )
        try:
            system_delta = float(d["cpu_stats"]["system_cpu_usage"]) - float(
                d["precpu_stats"]["system_cpu_usage"]
            )
        except KeyError:
            system_delta = 0.0
        if system_delta > 0.0:
            cpu_percent = cpu_delta / system_delta * 100.0 * cpu_count
        return cpu_percent

    def calculate_blkio_bytes(self, d):
        bytes_stats = d.get("blkio_stats", {}).get("io_service_bytes_recursive")
        if not bytes_stats:
            return 0, 0
        rd = wr = 0
        for s in bytes_stats:
            if s["op"] == "Read":
                rd += s["value"]
            elif s["op"] == "Write":
                wr += s["value"]
        return self.convert_size(rd, binary=False), self.convert_size(wr, binary=False)

    def calculate_network_bytes(self, d):
        networks = d.get("networks")
        if not networks:
            return 0, 0
        rx = tx = 0
        for data in networks.values():
            rx += data["rx_bytes"]
            tx += data["tx_bytes"]
        return self.convert_size(rx, binary=False), self.convert_size(tx, binary=False)

    def calculate_memory(self, d):
        memory = d.get("memory_stats")
        if not memory:
            return 0, 0
        return self.convert_size(
            memory.get("usage", 0), binary=True
        ), self.convert_size(memory.get("limit", 0), binary=True)



import json
import os
import re

from kubernetes import client, config

from .constants import NFS_PATH, NFS_SERVER, VOLUMES_ROOT


def split_and_clean_quotes(input_string):
    """
    Splits a string by whitespace, ignoring whitespace within single quotes,
    and removes the surrounding single quotes from the resulting quoted tokens.
    """
    # 1. Find all tokens (quoted or unquoted)
    pattern = r"'[^']+'|\S+"
    tokens = re.findall(pattern, input_string)

    # 2. Clean the quotes from the found tokens
    cleaned_tokens = []
    for token in tokens:
        # Check if the token starts and ends with a single quote
        if token.startswith("'") and token.endswith("'"):
            # Remove the first and last character (the quotes)
            cleaned_tokens.append(token[1:-1])
        else:
            # Keep unquoted tokens as they are
            cleaned_tokens.append(token)

    return cleaned_tokens


def tale_ingress(params: dict) -> None:
    host = f"{params['host']}.{params['domain']}"
    annotations = {
        "kubernetes.io/ingress.class": params["ingressClass"],
    }
    if params["ingressClass"] == "traefik":
        annotations[
            "traefik.ingress.kubernetes.io/router.middlewares"
        ] = f"{params['deploymentNamespace']}-ssl-header@kubernetescrd"
    elif params["ingressClass"] == "nginx":
        csp = f"frame-ancestors 'self' https://dashboard.{params['domain']};"
        annotations.update(
            {
                "nginx.ingress.kubernetes.io/proxy-read-timeout": "3600",
                "nginx.ingress.kubernetes.io/proxy-send-timeout": "3600",
                "nginx.ingress.kubernetes.io/proxy-http-version": "1.1",
                "nginx.ingress.kubernetes.io/proxy-body-size": "512M",
                "nginx.ingress.kubernetes.io/configuration-snippet": (
                    'more_clear_headers "Content-Security-Policy";\n'
                    f'add_header Content-Security-Policy "{csp}";\n'
                ),
            }
        )

    # Apply the Ingress
    config.load_incluster_config()
    api_instance = client.NetworkingV1Api()
    api_instance.create_namespaced_ingress(
        body=client.V1Ingress(
            metadata=client.V1ObjectMeta(
                name=f"tale-{params['host']}",
                labels={
                    "app": "WholeTale",
                    "component": params["deploymentName"],
                    "instanceId": params["instanceId"],
                },
                namespace=params["deploymentNamespace"],
                annotations=annotations,
            ),
            spec=client.V1IngressSpec(
                tls=[
                    client.V1IngressTLS(
                        hosts=[host],
                    )
                ],
                rules=[
                    client.V1IngressRule(
                        host=host,
                        http=client.V1HTTPIngressRuleValue(
                            paths=[
                                client.V1HTTPIngressPath(
                                    path="/",
                                    path_type="Prefix",
                                    backend=client.V1IngressBackend(
                                        service=client.V1IngressServiceBackend(
                                            name=params["deploymentName"],
                                            port=client.V1ServiceBackendPort(
                                                number=params["instancePort"],
                                            ),
                                        ),
                                    ),
                                ),
                            ],
                        ),
                    ),
                ],
            ),
        ),
        namespace=params["deploymentNamespace"],
    )


def tale_service(params):
    # Apply the service
    config.load_incluster_config()
    api_instance = client.CoreV1Api()
    api_instance.create_namespaced_service(
        body=client.V1Service(
            metadata=client.V1ObjectMeta(
                name=params["deploymentName"],
                labels={
                    "app": "WholeTale",
                    "component": params["deploymentName"],
                    "instanceId": params["instanceId"],
                },
            ),
            spec=client.V1ServiceSpec(
                selector={"app": "WholeTale", "instanceId": params["instanceId"]},
                ports=[
                    client.V1ServicePort(protocol="TCP", port=params["instancePort"])
                ],
            ),
        ),
        namespace=params["deploymentNamespace"],
    )


def compose_volumes(params):
    volumes = [
        client.V1Volume(name="data", empty_dir={}),
    ]

    if params["girderFSMountType"] == "direct":
        volumes.append(
            client.V1Volume(
                name=params["claimName"],
                persistent_volume_claim=client.V1PersistentVolumeClaimVolumeSource(
                    claim_name=params["claimName"],
                ),
            )
        )
        if NFS_SERVER:
            volumes += [
                client.V1Volume(
                    name="workspace",
                    nfs=client.V1NFSVolumeSource(
                        path=os.path.join(NFS_PATH, params["workspaceSubPath"]),
                        server=NFS_SERVER,
                    ),
                ),
                client.V1Volume(
                    name="home",
                    nfs=client.V1NFSVolumeSource(
                        path=os.path.join(NFS_PATH, params["homeSubPath"]),
                        server=NFS_SERVER,
                    ),
                ),
            ]
    else:
        volumes += [
            client.V1Volume(name="workspace", empty_dir={}),
            client.V1Volume(name="home", empty_dir={}),
        ]

    return volumes


def tale_deployment(params):
    config.load_incluster_config()

    volumes = compose_volumes(params)

    mounter_mounts = [
        client.V1VolumeMount(
            mount_path="/data",
            mount_propagation="Bidirectional",
            name="data",
        ),
    ]
    if params["girderFSMountType"] == "direct":
        mounter_mounts.append(
            client.V1VolumeMount(
                mount_path=VOLUMES_ROOT,
                name=params["claimName"],
                sub_path=params["claimSubPath"],
                read_only=False,
            ),
        )
    else:
        mounter_mounts += [
            client.V1VolumeMount(
                mount_path="/workspace",
                mount_propagation="Bidirectional",
                name="workspace",
                read_only=False,
            ),
            client.V1VolumeMount(
                mount_path="/home",
                mount_propagation="Bidirectional",
                name="home",
                read_only=False,
            ),
        ]

    instance_mounts = [
        client.V1VolumeMount(
            mount_path=os.path.join(params["mountPoint"], "data"),
            mount_propagation="HostToContainer",
            name="data",
        )
    ]

    if params["girderFSMountType"] == "direct" and not NFS_SERVER:
        instance_mounts += [
            client.V1VolumeMount(
                mount_path=os.path.join(params["mountPoint"], "workspace"),
                name=params["claimName"],
                sub_path=os.path.join(
                    params["claimSubPath"], params["workspaceSubPath"]
                ),
            ),
            client.V1VolumeMount(
                mount_path=os.path.join(params["mountPoint"], "home"),
                name=params["claimName"],
                sub_path=os.path.join(params["claimSubPath"], params["homeSubPath"]),
            ),
        ]
    else:
        instance_mounts = [
            client.V1VolumeMount(
                mount_path=os.path.join(params["mountPoint"], "workspace"),
                name="workspace",
            ),
            client.V1VolumeMount(
                mount_path=os.path.join(params["mountPoint"], "home"),
                name="home",
            ),
        ]

    # Apply the deployment
    command_array = split_and_clean_quotes(params["command"])
    api_instance = client.AppsV1Api()
    api_instance.create_namespaced_deployment(
        body=client.V1Deployment(
            metadata=client.V1ObjectMeta(
                name=params["deploymentName"],
                namespace=params["deploymentNamespace"],
                labels={
                    "app": "WholeTale",
                    "component": params["deploymentName"],
                    "instanceId": params["instanceId"],
                },
            ),
            spec=client.V1DeploymentSpec(
                selector=client.V1LabelSelector(
                    match_labels={
                        "app": "WholeTale",
                        "instanceId": params["instanceId"],
                    }
                ),
                strategy=client.V1DeploymentStrategy(type="Recreate"),
                template=client.V1PodTemplateSpec(
                    metadata=client.V1ObjectMeta(
                        labels={
                            "app": "WholeTale",
                            "instanceId": params["instanceId"],
                        }
                    ),
                    spec=client.V1PodSpec(
                        image_pull_secrets=[
                            client.V1LocalObjectReference(name="local-registry-secret")
                        ],
                        containers=[
                            client.V1Container(
                                name="instance",
                                command=[command_array[0]],
                                args=command_array[1:],
                                image=params["instanceImage"],
                                ports=[
                                    client.V1ContainerPort(
                                        container_port=params["instancePort"]
                                    )
                                ],
                                volume_mounts=instance_mounts,
                                resources=client.V1ResourceRequirements(
                                    limits={
                                        "memory": "4Gi",
                                        "cpu": "1",
                                    },
                                    requests={
                                        "memory": "4Gi",
                                        "cpu": "1",
                                    },
                                ),
                                env=[
                                    client.V1EnvVar(
                                        name="GIRDER_API_KEY",
                                        value=params["girderApiKey"],
                                    ),
                                    client.V1EnvVar(
                                        name="GIRDER_API_URL",
                                        value=params["girderApiUrl"],
                                    ),
                                    client.V1EnvVar(
                                        name="TMP_URL",
                                        value=f"{params['host']}.{params['domain']}",
                                    ),
                                ],
                            ),
                            client.V1Container(
                                name="mounter",
                                command=["tini", "--", "/bin/sleep", "infinity"],
                                image=params["mounterImage"],
                                lifecycle=client.V1Lifecycle(
                                    #    post_start=client.V1LifecycleHandler(
                                    #        _exec=client.V1ExecAction(
                                    #            command=["girderfs-mount"],
                                    #        )
                                    #    ),
                                    pre_stop=client.V1LifecycleHandler(
                                        _exec=client.V1ExecAction(
                                            command=["girderfs-umount"],
                                        )
                                    ),
                                ),
                                resources=client.V1ResourceRequirements(
                                    limits={
                                        "memory": "1Gi",
                                        "cpu": "1",
                                        "smarter-devices/fuse": 1,
                                    },
                                    requests={
                                        "memory": "1Gi",
                                        "cpu": "0.5",
                                        "smarter-devices/fuse": 1,
                                    },
                                ),
                                env=[
                                    client.V1EnvVar(
                                        name="GIRDERFS_DEF",
                                        value=json.dumps(params["girderfsDef"]),
                                    ),
                                    client.V1EnvVar(
                                        name="WT_VOLUMES_PATH",
                                        value=VOLUMES_ROOT,
                                    ),
                                ],
                                security_context=client.V1SecurityContext(
                                    allow_privilege_escalation=True,
                                    capabilities=client.V1Capabilities(
                                        add=["SYS_ADMIN"]
                                    ),
                                    privileged=True,
                                ),
                                volume_mounts=mounter_mounts,
                            ),
                        ],
                        volumes=volumes,
                    ),
                ),
            ),
        ),
        namespace=params["deploymentNamespace"],
    )



import json
import logging
import os
import time
from urllib.parse import urlparse

import girder_client

from .constants import BUILD_TALE_IMAGE_STEP_TOTAL, InstanceStatus, TaleStatus
from .lib.zenodo import ZenodoPublishProvider
from .r2d import ImageBuilder


class TasksBase:
    def create_volume(self, task, instance_id: str):
        raise NotImplementedError()

    def launch_container(self, task, service_info):
        raise NotImplementedError()

    def update_container(self, task, instanceId, digest=None):
        raise NotImplementedError()

    def shutdown_container(self, task, instanceId):
        raise NotImplementedError()

    def remove_volume(self, task, instanceId):
        raise NotImplementedError()

    def build_tale_image(self, task, tale_id, force=False):
        """
        Build docker image from Tale workspace using repo2docker and push to Whole Tale registry.
        """
        logging.info("Building image for Tale %s", tale_id)

        task.job_manager.updateProgress(
            message="Building image",
            total=BUILD_TALE_IMAGE_STEP_TOTAL,
            current=1,
            forceFlush=True,
        )

        tic = time.time()
        tale = task.girder_client.get("/tale/%s" % tale_id)
        while tale["status"] != TaleStatus.READY:
            time.sleep(2)
            tale = task.girder_client.get("/tale/{_id}".format(**tale))
            if tale["status"] == TaleStatus.ERROR:
                raise ValueError("Cannot build image for a Tale in error state.")
            if time.time() - tic > 5 * 60.0:
                raise ValueError(
                    "Cannot build image. Tale preparing for more than 5 minutes."
                )

        last_build_time = -1
        try:
            last_build_time = tale["imageInfo"]["last_build"]
        except KeyError:
            pass

        logging.info("Last build time {}".format(last_build_time))
        image_builder = ImageBuilder(task.girder_client, tale=tale)
        image_builder.pull_r2d()

        tag = image_builder.get_tag(force=force)

        logging.info("Computed tag: %s (taleId:%s)", tag, tale_id)

        # Use the current time as the image build time and tag
        build_time = int(time.time())

        # Check if image already exists
        print("Checking if image exists...")
        print(f"Forced build: {force}")
        print(f"Last build time: {last_build_time}")
        print(f"image_builder.cached_image(tag): {image_builder.cached_image(tag)}")
        if not force and (image := image_builder.cached_image(tag)):
            print("Cached image exists for this Tale. Skipping build.")
            task.job_manager.updateProgress(
                message="Tale not modified, no need to build",
                total=BUILD_TALE_IMAGE_STEP_TOTAL,
                current=BUILD_TALE_IMAGE_STEP_TOTAL,
                forceFlush=True,
            )
            return {
                "image_digest": f"{image['name']}:{image['tag']}@{image['digest']}",
                "repo2docker_version": image_builder.container_config.repo2docker_version,
                "last_build": last_build_time,
            }

        print("Forcing build.")

        # Prepare build context
        ret, _ = image_builder.run_r2d(tag, task=task)
        if task.canceled:
            task.request.chain = None
            logging.info("Build canceled.")
            return

        if ret["StatusCode"] != 0:
            # repo2docker build failed
            print(ret)
            raise ValueError("Error building tale {}".format(tale_id))

        # Push the image to the registry
        logging.info("Pushing image %s", tag)
        image_builder.push_image(tag)
        logging.info("Image pushed")

        # Get the built image digest
        logging.info("Getting image from cache...")
        image = image_builder.cached_image(tag)
        logging.info("Image: %s", image)

        task.job_manager.updateProgress(
            message="Image build succeeded",
            total=BUILD_TALE_IMAGE_STEP_TOTAL,
            current=BUILD_TALE_IMAGE_STEP_TOTAL,
            forceFlush=True,
        )

        logging.info(
            f"Successfully built image {image['name']}:{image['tag']} ({image['digest']})"
        )

        # Image digest used by updateBuildStatus handler
        return {
            "image_digest": f"{image['name']}:{image['tag']}@{image['digest']}",
            "repo2docker_version": image_builder.container_config.repo2docker_version,
            "last_build": build_time,
        }

    def publish(self, task, tale_id, token, version_id, repository=None, draft=False):
        """
        Publish a tale.

        :param tale_id: The tale id
        :param token: An access token for a given repository.
        :param version_id: The version of the Tale being published
        :param repository: Target repository.
        :param draft: If True, don't mint DOI.
        :type tale_id: str
        :type token: obj
        :type repository: str
        :type draft: boolean
        """

        provider_name = token["provider"].lower()
        if provider_name == "zenodo":
            provider = ZenodoPublishProvider(
                task.girder_client,
                tale_id,
                token,
                version_id,
                draft=draft,
                job_manager=task.job_manager,
            )
        else:
            raise ValueError("Unsupported publisher ({})".format(token["provider"]))

        if (
            provider.published
            and provider.publication_info.get("versionId") == version_id
        ):
            raise ValueError(
                f"This version of the Tale ({version_id}) has already been published."
            )
        provider.publish()
        return provider.publication_info

    def import_tale(self, task, lookup_kwargs, tale, spawn=True):
        """Create a Tale provided a url for an external data and an image Id.

        Currently, this task only handles importing raw data. In the future, it
        should also allow importing serialized Tales.
        """
        if spawn:
            total = 4
        else:
            total = 3

        if spawn:
            try:
                instance = task.girder_client.post(
                    "/instance", parameters={"taleId": tale["_id"]}
                )
            except girder_client.HttpError as resp:
                try:
                    message = json.loads(resp.responseText).get("message", "")
                except json.JSONDecodeError:
                    message = str(resp)
                errormsg = "Unable to create instance. Server returned {}: {}"
                errormsg = errormsg.format(resp.status, message)

        def set_tale_error_status():
            task.girder_client.put(
                "/tale/{_id}".format(**tale),
                json={
                    "status": TaleStatus.ERROR,
                    "imageId": str(tale["imageId"]),
                    "public": tale["public"],
                },
            )

        task.job_manager.updateProgress(
            message="Gathering basic info about the dataset", total=total, current=1
        )
        dataId = lookup_kwargs.pop("dataId")
        try:
            parameters = dict(dataId=json.dumps(dataId))
            parameters.update(lookup_kwargs)
            dataMap = task.girder_client.get(
                "/repository/lookup", parameters=parameters
            )
        except girder_client.HttpError as resp:
            try:
                message = json.loads(resp.responseText).get("message", "")
            except json.JSONDecodeError:
                message = str(resp)
            errormsg = 'Unable to register "{}". Server returned {}: {}'
            errormsg = errormsg.format(dataId[0], resp.status, message)
            set_tale_error_status()
            raise ValueError(errormsg)

        if not dataMap:
            errormsg = 'Unable to register "{}". Source is not supported'
            errormsg = errormsg.format(dataId[0])
            set_tale_error_status()
            raise ValueError(errormsg)

        task.job_manager.updateProgress(
            message="Registering the dataset in Whole Tale", total=total, current=2
        )
        parameters = {"dataMap": json.dumps(dataMap)}
        try:
            parameters["base_url"] = lookup_kwargs.pop("base_url")
        except KeyError:
            pass
        task.girder_client.post("/dataset/register", parameters=parameters)

        # Currently, we register resources in two different ways:
        #  1. DOIs (coming from Globus, Dataverse, DataONE, etc) create a root
        #     folder in the Catalog, that's named exactly the same as dataset.
        #  2. HTTP(S) files are registered into Catalog using a nested structure
        #     based on their url (see whole-tale/girder_wholetale#266)
        #  Knowing that, let's try to find the newly registered data by path.
        catalog_path = "/collection/WholeTale Catalog/WholeTale Catalog"
        if dataMap[0]["repository"].lower().startswith("http"):
            url = urlparse(dataMap[0]["dataId"])
            path = os.path.join(catalog_path, url.netloc, url.path[1:])
        else:
            path = os.path.join(catalog_path, dataMap[0]["name"])

        resource = task.girder_client.get("/resource/lookup", parameters={"path": path})
        if not resource:
            errormsg = "Registration of {} failed. Aborting!".format(
                dataMap[0]["dataId"]
            )
            set_tale_error_status()
            raise ValueError(errormsg)

        tale["dataSet"] = [
            {
                "mountPath": resource["name"],
                "itemId": resource["_id"],
                "_modelType": resource["_modelType"],
            }
        ]
        tale = task.girder_client.put(
            "/tale/{_id}".format(**tale),
            json={
                "dataSet": tale["dataSet"],
                "imageId": str(tale["imageId"]),
                "public": tale["public"],
                "status": TaleStatus.READY,
            },
        )

        if spawn:
            task.job_manager.updateProgress(
                message="Creating a Tale container", total=total, current=3
            )
            while instance["status"] == InstanceStatus.LAUNCHING:
                # TODO: Timeout? Raise error?
                time.sleep(1)
                instance = task.girder_client.get("/instance/{_id}".format(**instance))
        else:
            instance = None

        task.job_manager.updateProgress(
            message="Tale is ready!", total=total, current=total
        )
        # TODO: maybe filter results?
        return {"tale": tale, "instance": instance}

    def recorded_run(self, task, run_id, tale_id, entrypoint):
        raise NotImplementedError()

    def check_on_run(self, run_state):
        raise NotImplementedError()

    def cleanup_run(self, task, run_id):
        raise NotImplementedError()



"""Whole Tale publisher provider for Zenodo."""

import datetime
import json
import logging
import requests
import tempfile
from urllib.parse import urlencode, urlparse

from markdown import Markdown
from lxml.html.clean import Cleaner

from .publish_provider import PublishProvider
from ..utils import DEPLOYMENT


_ZENODO_ALLOWED_TAGS = {
    "a",
    "p",
    "br",
    "blockquote",
    "strong",
    "b",
    "u",
    "i",
    "em",
    "ul",
    "ol",
    "li",
    "sub",
    "sup",
    "div",
    "strike",
}
_ZENODO_ACCEPTED_RELATIONS = {
    "isCitedBy",
    "cites",
    "isSupplementTo",
    "isSupplementedBy",
    "isContinuedBy",
    "continues",
    "hasMetadata",
    "isMetadataFor",
    "isNewVersionOf",
    "isPreviousVersionOf",
    "isPartOf",
    "hasPart",
    "isReferencedBy",
    "references",
    "isDocumentedBy",
    "documents",
    "isCompiledBy",
    "compiles",
    "isVariantFormOf",
    "isOrignialFormOf",
    "isIdenticalTo",
    "isReviewedBy",
    "reviews",
    "isDerivedFrom",
    "isSourceOf"
}
_ACTION_CODES = {"publish": 202, "newversion": 201}


class ZenodoPublishProvider(PublishProvider):
    """Whole Tale publisher provider for Zenodo."""

    total_updates = 3
    current_update = 0

    def update_progress(self, message=None):
        """Send update and increment update counter."""
        self.current_update += 1
        self.job_manager.updateProgress(
            total=self.total_updates, current=self.current_update, message=message
        )

    @staticmethod
    def _render_description(description):
        """Convert Markdown to HTML accepted by Zenodo."""
        parser = Markdown()
        html = parser.convert(description)
        html_cleaner = Cleaner(
            allow_tags=_ZENODO_ALLOWED_TAGS, remove_unknown_tags=False
        )
        return html_cleaner.clean_html(html)

    def deposit_tale(self, deposition):
        """Given a Zenodo deposition, upload tarball with a Tale."""

        self.update_progress(message="Uploading data to " + self.resource_server)

        stream = self.gc.sendRestRequest(
            "get",
            "tale/{}/export".format(self.tale["_id"]),
            parameters={"taleFormat": "bagit", "versionId": self.version_id},
            stream=True,
            jsonResp=False,
        )
        with tempfile.NamedTemporaryFile(suffix=".zip") as tmp:

            # Write the zip file
            for chunk in stream.iter_content(chunk_size=65536):
                tmp.write(chunk)
            tmp.seek(0)

            r = self.request(
                "/api/deposit/depositions/{}/files".format(deposition["id"]),
                method="POST",
                data={"name": "{}.zip".format(self.version_id)},
                files={"file": tmp},
            )

            try:
                r.raise_for_status()
            except requests.HTTPError as exc:
                msg = "Failed to upload to a deposition (id:{}).".format(
                    deposition["id"]
                )
                msg += " Server returned: " + str(exc)
                if r.status_code > 399 and r.status_code < 500:
                    msg += "\n" + json.dumps(r.json(), sort_keys=True, indent=4) + "\n"
                logging.warning(msg)
                raise ValueError(msg)
            logging.debug(
                "[zenodo:deposit_tale] POST depositions/{}/files = {}".format(
                    deposition["id"], r.json()
                )
            )

    def get_tale_metadata(self):
        """Convert the Tale metadata to a Zenodo metadata."""

        keywords = {"Tale"}
        if self.manifest.get("schema:keywords"):
            keywords.add(self.manifest["schema:keywords"].title())

        def first_letter_lower(s):
            return s[:1].lower() + s[1:] if s else ""

        related_identifiers = []
        for related_id in self.tale.get("datacite:relatedIdentifiers", []):
            relation = related_id["datacite:relatedIdentifier"]
            relation_type = first_letter_lower(relation["datacite:relationType"][3:])
            if relation_type not in _ZENODO_ACCEPTED_RELATIONS:
                continue
            related_identifiers.append(
                {"relation": relation_type, "identifier": relation["@id"]}
            )
        related_identifiers += [
            {"relation": "cites", "identifier": ds["schema:identifier"]}
            for ds in self.manifest.get("wt:usesDataset", [])
        ]
        # Remove duplicates
        related_identifiers = [
            json.loads(rel_id)
            for rel_id in {json.dumps(_, sort_keys=True) for _ in related_identifiers}
        ]

        return {
            "metadata": {
                "upload_type": "publication",
                "publication_type": "other",
                "title": self.manifest["schema:name"],
                "creators": [
                    {
                        "name": "{}, {}".format(
                            author["schema:familyName"], author["schema:givenName"]
                        ),
                        "orcid": author["@id"].split("/")[-1],
                    }
                    for author in self.manifest["schema:author"]
                ],
                "description": self._render_description(
                    self.manifest["schema:description"]
                ),
                "keywords": list(keywords),
                "access_right": "open",
                "license": self.tale["licenseSPDX"].lower(),
                "prereserve_doi": True,
                "related_identifiers": related_identifiers,
                "references": [
                    citation for citation in self.tale.get("dataSetCitation", [])
                ],
            }
        }

    def create_deposition(self):
        """Create a deposition @ Zenodo."""
        zenodo_metadata = self.get_tale_metadata()
        logging.debug(
            "[zenodo:create_deposition] metadata = {}".format(zenodo_metadata)
        )

        self.update_progress(message="Creating a deposition at " + self.resource_server)

        r = self.request(
            "/api/deposit/depositions",
            method="POST",
            data=json.dumps(zenodo_metadata),
            headers={"Content-Type": "application/json"},
        )

        try:
            r.raise_for_status()
        except requests.HTTPError as exc:
            msg = "Failed to create a deposition. Server returned: " + str(exc)
            if r.status_code > 399 and r.status_code < 500:
                msg += "\n" + json.dumps(r.json(), sort_keys=True, indent=4) + "\n"
            logging.warning(msg)
            raise ValueError(msg)
        logging.debug("[zenodo:create_deposition] deposition = {}".format(r.json()))
        return r.json()

    def _zenodo_action(self, deposition, action):
        r = self.request(
            "/api/deposit/depositions/{dep_id}/actions/{action}".format(
                dep_id=deposition["id"], action=action
            ),
            method="POST",
        )
        assert r.status_code == _ACTION_CODES[action]
        logging.debug("[zenodo:_action_{}] deposition = {}".format(action, r.json()))
        return r.json()

    def publish_deposition(self, deposition):
        """Publish publish, I mean REALLY publish."""
        return self._zenodo_action(deposition, "publish")

    def create_new_version(self, deposition):
        """Create a new version of an exisiting deposition."""
        current_deposition = self._zenodo_action(deposition, "newversion")
        # From zenodo docs:
        # NOTE: The response body of this action is NOT the new version deposit,
        # but the original resource. The new version deposition can be accessed
        # through the "latest_draft" under "links" in the response body.
        return self.retrieve_deposition(current_deposition["links"]["latest_draft"])

    def update_deposition(self, deposition):
        """Update deposition, assumes that metadata is changed."""
        r = self.request(
            "/api/deposit/depositions/{}".format(deposition["id"]),
            method="PUT",
            data=json.dumps({"metadata": deposition["metadata"]}),
            headers={"Content-Type": "application/json"},
        )

        try:
            r.raise_for_status()
        except requests.HTTPError as exc:
            msg = "Failed to update the deposition (id={}).".format(deposition["id"])
            msg += " Server returned: " + str(exc)
            if r.status_code > 399 and r.status_code < 500:
                msg += "\n" + json.dumps(r.json(), sort_keys=True, indent=4) + "\n"
            logging.warning(msg)
            raise ValueError(msg)
        logging.debug("[zenodo:update_deposition] deposition = {}".format(r.json()))
        return r.json()

    def retrieve_deposition(self, deposition_id):
        """Get an existing deposition provided id or url."""
        if deposition_id.startswith("http"):
            url = urlparse(deposition_id).path
        else:
            url = "/api/deposit/depositions/{}".format(deposition_id)

        r = self.request(
            url, method="GET", headers={"Content-Type": "application/json"}
        )

        try:
            r.raise_for_status()
        except requests.HTTPError as exc:
            msg = "Failed to get the deposition (id={}).".format(deposition_id)
            msg += " Server returned: " + str(exc)
            if r.status_code > 399 and r.status_code < 500:
                msg += "\n" + json.dumps(r.json(), sort_keys=True, indent=4) + "\n"
            logging.warning(msg)
            raise ValueError(msg)
        logging.debug("[zenodo:retrieve_deposition] deposition = {}".format(r.json()))
        return r.json()

    def remove_files(self, deposition):
        for fileobj in deposition["files"]:
            r = self.request(
                "/api/deposit/depositions/{dep_id}/files/{file_id}".format(
                    dep_id=deposition["id"], file_id=fileobj["id"]
                ),
                method="DELETE",
            )

            try:
                r.raise_for_status()
            except requests.HTTPError as exc:
                msg = "Failed to remove file (id={}) from the deposition (id={}).".format(
                    fileobj["id"], deposition["id"]
                )
                msg += " Server returned: " + str(exc)
                if r.status_code > 399 and r.status_code < 500:
                    msg += "\n" + json.dumps(r.json(), sort_keys=True, indent=4) + "\n"
                logging.warning(msg)
                raise ValueError(msg)

    def publish(self):
        """Publish the specified Tale to Zenodo."""
        if self.published:
            old_deposition = self.retrieve_deposition(
                self.publication_info["repository_id"]
            )
            deposition = self.create_new_version(old_deposition)

            # NOTE: If Tale was already published the only way to "update" the
            # payload is to delete the file and upload a new file.
            self.remove_files(deposition)

            deposition.update(self.get_tale_metadata())
            deposition = self.update_deposition(deposition)

            # TODO: Discard version if something goes wrong...
        else:
            deposition = self.create_deposition()
        self.publish_version(deposition)

    def publish_version(self, deposition):
        # Zenodo is nice and gives up preserved DOI at this point.
        # We should update the Tale before publishing so that
        # the manifest reflects that, shouldn't we?
        # TODO: update Tale's manifest
        try:
            # TODO: not sure why, but prereserve_doi doesn't work in sandbox occasionally
            doi = deposition["metadata"]["prereserve_doi"]["doi"]
        except KeyError:
            doi = None

        # Add a self reference pointing to WT
        msg = (
            "Run this Tale on Whole Tale by clicking "
            '<a href="{girder_url}/api/v1/integration/zenodo?{query}">here</a>.'
        ).format(
            girder_url=DEPLOYMENT.girder_url,
            query=urlencode({"doi": doi, "resource_server": self.resource_server}),
        )
        deposition["metadata"]["notes"] = msg
        try:
            deposition = self.update_deposition(deposition)
        except Exception as exc:
            # We should delete deposition if something goes wrong
            self.rollback(deposition)
            raise exc

        try:
            self.deposit_tale(deposition)
        except Exception as exc:
            # We should delete deposition if something goes wrong
            self.rollback(deposition)
            raise exc

        # TODO: since it's a two step process and here's the moment when we have a
        # "camera-ready" Tale deposited and ready for final click, it'd be stupendous
        # if we had a UI component...
        if self.draft:
            published_url = deposition["links"]["self"]
        else:
            # Rollback is not possible afterwards...
            deposition = self.publish_deposition(deposition)
            doi = deposition["doi"]
            published_url = deposition["links"]["record_html"]

        self.update_progress(
            message="Your Tale has successfully been published to " + published_url
        )

        if doi and not doi.startswith("doi:"):
            doi = "doi:{}".format(doi)

        publish_info = {
            "pid": doi,
            "uri": published_url,
            "date": datetime.datetime.utcnow().isoformat(),
            "repository_id": str(deposition["id"]),
            "repository": self.resource_server,
            "versionId": self.version_id,
        }
        if self.published:
            self.tale["publishInfo"][self._published_info_index].update(publish_info)
        else:
            self.tale["publishInfo"].append(publish_info)

        # Update the Tale with a reference to the published resource
        try:
            self.gc.put("tale/{}".format(self.tale["_id"]), json=self.tale)
        except Exception as e:
            msg = "Error updating Tale {}".format(str(e))
            logging.warning(msg)
            raise ValueError(msg)

    def rollback(self, deposition):
        """Delete an unpublished deposition."""
        r = self.request(
            "/api/deposit/depositions/{}".format(deposition["id"]), method="DELETE"
        )
        logging.warning("Deleting returned : {}".format(r.status_code))
        assert r.status_code == 204  # Docs say it should return 201...

    def request(self, path, method="GET", data=None, files=None, headers=None):
        """Wrap commonly used requests to pass header and create url."""
        headers = headers or {}
        headers.update({"Authorization": "Bearer {}".format(self.access_token)})
        url = "https://{}{}".format(self.resource_server, path)

        if method.upper() == "GET":
            return requests.get(url, headers=headers)
        elif method.upper() == "PUT":
            return requests.put(url, data=data, headers=headers)
        elif method.upper() == "POST":
            return requests.post(url, data=data, files=files, headers=headers)
        elif method.upper() == "DELETE":
            return requests.delete(url, headers=headers)



from setuptools import setup, find_packages

setup(
    name="gwvolman",
    version="2.1.3",
    description="WholeTale Girder Volume Manager",
    author="Kacper Kowalik",
    author_email="xarthisius.kk@gmail.com",
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Programming Language :: Python",
    ],
    entry_points={
        "girder_worker_plugins": [
            "gwvolman = gwvolman:GWVolumeManagerPlugin",
        ]
    },
    install_requires=[
        "girder-client",
        "girder-worker>=5.0.0a5.dev0",
        "kubernetes",
        "docker>=2.3.0",
        "requests",
        "markdown",
        "lxml_html_clean",
        "pystache",
        "celery[redis]>5",
        "python-dateutil",
    ],
    packages=find_packages(),
    zip_safe=False,
)



import docker
from datetime import datetime, timedelta
import os
import time
import logging
import json

from .constants import (
    CREATE_VOLUME_STEP_TOTAL,
    LAUNCH_CONTAINER_STEP_TOTAL,
    RECORDED_RUN_STEP_TOTAL,
    UPDATE_CONTAINER_STEP_TOTAL,
)
from .utils import (
    new_user,
    _get_api_key,
    _get_container_config,
    _launch_container,
    _get_user_and_instance,
    _recorded_run,
    stop_container,
)
from .fs_container import FSContainer
from .tasks_base import TasksBase
from .constants import GIRDER_API_URL, RunStatus, VOLUMES_ROOT

from .r2d import DockerImageBuilder


class DockerTasks(TasksBase):
    def create_volume(self, task, instance_id, mounts=None):
        """Create a mountpoint and compose WT-fs."""
        user, instance = _get_user_and_instance(task.girder_client, instance_id)
        tale = task.girder_client.get("/tale/{taleId}".format(**instance))

        task.job_manager.updateProgress(
            message="Creating volume",
            total=CREATE_VOLUME_STEP_TOTAL,
            current=1,
            forceFlush=True,
        )

        vol_name = "%s_%s_%s" % (tale["_id"], user["login"], new_user(6))
        fs_sidecar = FSContainer.start_container(vol_name)
        if mounts is None:
            mounts = [
                {
                    "type": "data",
                    "protocol": "girderfs",
                    "location": "data",
                },
                {
                    "type": "home",
                    "protocol": "bind",
                    "location": "home",
                },
                {
                    "type": "workspace",
                    "protocol": "bind",
                    "location": "workspace",
                },
                {
                    "type": "versions",
                    "protocol": "girderfs",
                    "location": "versions",
                },
                {
                    "type": "runs",
                    "protocol": "girderfs",
                    "location": "runs",
                },
            ]

        payload = {
            "mounts": mounts,
            "taleId": tale["_id"],
            "userId": user["_id"],
            "girderApiUrl": GIRDER_API_URL,
            "girderApiKey": _get_api_key(task.girder_client),
            "girderToken": task.girder_client.token,
            "root": vol_name,
        }
        print(json.dumps(payload))
        FSContainer.mount(fs_sidecar, payload)
        task.job_manager.updateProgress(
            message="Volume created",
            total=CREATE_VOLUME_STEP_TOTAL,
            current=CREATE_VOLUME_STEP_TOTAL,
            forceFlush=True,
        )
        print("WT Filesystem created successfully.")

        cli = docker.from_env()
        return dict(
            nodeId=cli.info()["Swarm"]["NodeID"],
            fscontainerId=fs_sidecar.id,
            volumeName=vol_name,
            instanceId=instance_id,
            taleId=tale["_id"],
        )

    def launch_container(self, task, service_info):
        """Launch a container using a Tale object."""
        user, instance = _get_user_and_instance(
            task.girder_client, service_info["instanceId"]
        )
        tale = task.girder_client.get(f"/tale/{service_info['taleId']}")

        task.job_manager.updateProgress(
            message="Starting container",
            total=LAUNCH_CONTAINER_STEP_TOTAL,
            current=1,
            forceFlush=True,
        )

        print("Launching container for a Tale...")
        if "imageInfo" not in tale:
            # Wait for image to be built
            tic = time.time()
            timeout = 180.0
            time_interval = 5

            while time.time() - tic < timeout:
                tale = task.girder_client.get("/tale/{taleId}".format(**instance))
                if "imageInfo" in tale and "digest" in tale["imageInfo"]:
                    break
                msg = f"Waiting for image build to complete. ({time_interval}s)"
                logging.info(msg)
                print(msg)
                time.sleep(5)

        container_config = _get_container_config(task.girder_client, tale)
        service, attrs = _launch_container(
            service_info, container_config, task.girder_client
        )
        print(
            f"Started a container using volume: {service_info['volumeName']} "
            f"on node: {service_info['nodeId']}"
        )

        # wait until task is started
        tic = time.time()
        timeout = 300.0
        started = False

        print("Waiting for the environment to be accessible...")
        while time.time() - tic < timeout:
            try:
                status = service.tasks()[0]["Status"]

                if status["State"] in {"failed", "rejected"}:
                    raise ValueError("Failed to start environment: %s" % status["Err"])
                elif status["State"] == "running":
                    started = True
                    break

            except IndexError:
                started = False

            time.sleep(0.2)

        if not started:
            raise ValueError("Tale did not start before timeout exceeded")

        print("Environment is up and running.")
        task.job_manager.updateProgress(
            message="Container started",
            total=LAUNCH_CONTAINER_STEP_TOTAL,
            current=LAUNCH_CONTAINER_STEP_TOTAL,
            forceFlush=True,
        )

        service_info.update(attrs)
        service_info["name"] = service.name
        return service_info

    def update_container(self, task, instanceId, digest=None):
        user, instance = _get_user_and_instance(task.girder_client, instanceId)

        cli = docker.from_env(version="1.28")
        if "containerInfo" not in instance:
            return
        containerInfo = instance["containerInfo"]  # VALIDATE
        try:
            service = cli.services.get(containerInfo["name"])
        except docker.errors.NotFound:
            logging.info("Service not present [%s].", containerInfo["name"])
            return

        task.job_manager.updateProgress(
            message="Restarting the Tale with a new image",
            total=UPDATE_CONTAINER_STEP_TOTAL,
            current=1,
            forceFlush=True,
        )

        # Don't try to restart if the image hasn't changed
        try:
            previous_image = service.attrs["Spec"]["TaskTemplate"]["ContainerSpec"][
                "Image"
            ]
        except KeyError:
            previous_image = ""

        if previous_image == digest:
            task.job_manager.updateProgress(
                message="Image has not changed",
                total=UPDATE_CONTAINER_STEP_TOTAL,
                current=UPDATE_CONTAINER_STEP_TOTAL,
            )
            return {"image_digest": digest}

        try:
            # NOTE: Only "image" passed currently, but this can be easily extended
            logging.info("Restarting container [%s].", service.name)
            service.update(image=digest)
            logging.info(
                "Restart command has been sent to Container [%s].", service.name
            )
        except Exception as e:
            logging.error(
                "Unable to send restart command to container [%s]: %s", service.id, e
            )

        updated = False
        expired = False
        timeout = datetime.now() + timedelta(minutes=3)
        while not (updated or expired or task.canceled):
            service = cli.services.get(containerInfo["name"])

            try:
                state = service.attrs["UpdateStatus"]["State"]
            except KeyError:
                state = ""

            if state == "paused":
                raise RuntimeError(
                    'Restarting the Tale failed with "{}"'.format(
                        service.attrs["UpdateStatus"]["Message"]
                    )
                )

            updated = state == "completed"
            expired = datetime.now() > timeout
            time.sleep(1.0)

        if task.canceled:
            raise RuntimeError("Tale restart cancelled")
        elif expired:
            raise RuntimeError("Tale update timed out")

        task.job_manager.updateProgress(
            message="Tale restarted with the new image",
            total=UPDATE_CONTAINER_STEP_TOTAL,
            current=UPDATE_CONTAINER_STEP_TOTAL,
        )

        return {"image_digest": digest}

    def shutdown_container(self, task, instanceId):
        """Shutdown a running Tale."""
        user, instance = _get_user_and_instance(task.girder_client, instanceId)

        cli = docker.from_env(version="1.28")
        if "containerInfo" not in instance:
            return
        containerInfo = instance["containerInfo"]  # VALIDATE
        try:
            service = cli.services.get(containerInfo["name"])
        except docker.errors.NotFound:
            logging.info("Service not present [%s].", containerInfo["name"])
            return

        try:
            logging.info("Releasing container [%s].", service.name)
            service.remove()
            logging.info("Container [%s] has been released.", service.name)
        except Exception as e:
            logging.error("Unable to release container [%s]: %s", service.id, e)

    def remove_volume(self, task, instanceId):
        """Unmount WT-fs and remove mountpoint."""
        logging.info("Stopping FS container for instance %s", instanceId)
        user, instance = _get_user_and_instance(task.girder_client, instanceId)

        if "containerInfo" not in instance:
            logging.warning("No containerInfo for instance %s", instanceId)
            return
        containerInfo = instance["containerInfo"]  # VALIDATE
        FSContainer.stop_container(containerInfo["fscontainerId"])
        logging.info("FS container %s stopped", containerInfo["fscontainerId"])

    def recorded_run(self, task, run_id, tale_id, entrypoint):
        """Start a recorded run for a tale version"""
        run = task.girder_client.get(f"/run/{run_id}")
        state = RecordedRunCleaner(run, task.girder_client)
        tale = task.girder_client.get(
            f"/tale/{tale_id}/restore", parameters={"versionId": run["runVersionId"]}
        )
        user = task.girder_client.get("/user/me")
        image_builder = DockerImageBuilder(task.girder_client, tale=tale)

        def set_run_status(run, status):
            task.girder_client.patch(
                "/run/{_id}/status".format(**run), parameters={"status": status}
            )

        # UNKNOWN = 0 STARTING = 1 RUNNING = 2 COMPLETED = 3 FAILED = 4 CANCELLED = 5
        set_run_status(run, RunStatus.STARTING)

        task.job_manager.updateProgress(
            message="Preparing volumes",
            total=RECORDED_RUN_STEP_TOTAL,
            current=1,
            forceFlush=True,
        )

        # Create Docker volume
        vol_name = "%s_%s_%s" % (run_id, user["login"], new_user(6))
        fs_sidecar = FSContainer.start_container(vol_name)
        payload = {
            "mounts": [
                {
                    "type": "data",
                    "protocol": "girderfs",
                    "location": "data",
                },
                {
                    "type": "run",
                    "protocol": "bind",
                    "location": "workspace",
                },
            ],
            "girderApiUrl": GIRDER_API_URL,
            "girderApiKey": _get_api_key(task.girder_client),
            "root": vol_name,
            "taleId": tale["_id"],
            "runId": run["_id"],
            "userId": user["_id"],
        }
        FSContainer.mount(fs_sidecar, payload)
        state.volume_created = vol_name

        # Build the image for the run
        task.job_manager.updateProgress(
            message="Building image",
            total=RECORDED_RUN_STEP_TOTAL,
            current=2,
            forceFlush=True,
        )

        # Setup image tag
        tag = image_builder.get_tag()
        logging.info(
            "Computed tag: %s (taleId:%s, versionId:%s)",
            tag,
            tale_id,
            run["runVersionId"],
        )

        # Build currently assumes tmp directory, in this case mount the run workspace
        container_config = image_builder.container_config

        if task.canceled:
            state.cleanup()
            return

        try:
            if not image_builder.cached_image(tag):
                print("Building image for recorded run " + tag)
                ret, _ = image_builder.run_r2d(tag)
                if task.canceled:
                    state.cleanup()
                    return
                if ret["StatusCode"] != 0:
                    raise ValueError(
                        "Image build failed for recorded run {}".format(run_id)
                    )
                image_builder.push_image(tag)

            task.job_manager.updateProgress(
                message="Recording run",
                total=RECORDED_RUN_STEP_TOTAL,
                current=3,
                forceFlush=True,
            )

            set_run_status(run, RunStatus.RUNNING)
            container_name = f"rrun-{new_user(8)}"

            task.girder_client.addMetadataToFolder(
                run["_id"],
                {
                    "container_name": container_name,
                    "volume_created": state.volume_created,
                    "node_id": image_builder.dh.cli.info()["Swarm"]["NodeID"],
                },
            )

            mountpoint = os.path.join(VOLUMES_ROOT, "mountpoints", state.volume_created)

            _recorded_run(
                image_builder.dh.cli,
                mountpoint,
                container_config,
                tag,
                entrypoint,
                container_name,
                task=task,
            )
            if task.canceled:
                state.cleanup()
                return

            set_run_status(run, RunStatus.COMPLETED)
            task.job_manager.updateProgress(
                message="Finished recorded run",
                total=RECORDED_RUN_STEP_TOTAL,
                current=4,
                forceFlush=True,
            )
        except Exception as exc:
            logging.error(exc, exc_info=True)
            raise
        finally:
            state.cleanup(False)

    def check_on_run(self, run_state):
        cli = docker.from_env(version="1.28")
        try:
            container = cli.containers.get(run_state["container_name"])
            return container.status == "running"
        except docker.errors.NotFound:
            return False

    def cleanup_run(self, task, run_id):
        run = task.girder_client.get(f"/run/{run_id}")
        state = RecordedRunCleaner(run, task.girder_client)
        state.volume_created = run["meta"].get("volume_created")
        state.container_name = run["meta"].get("container_name")
        state.cleanup(canceled=False)
        state.set_run_status(RunStatus.FAILED)
        if job_id := run["meta"]["jobId"]:
            task.girder_client.put(f"/job/{job_id}", parameters={"status": 4})


class RecordedRunCleaner:
    volume_created = None
    container_name = None

    def __init__(self, run, gc):
        self.gc = gc
        self.run = run
        self.docker_cli = docker.from_env(version="1.28")

    def set_run_status(self, status):
        self.gc.patch(
            "/run/{_id}/status".format(**self.run), parameters={"status": status}
        )

    def cleanup(self, canceled=True):
        if self.container_name:
            container = self.docker_cli.containers.get(self.container_name)
            stop_container(container)

        if self.volume_created:
            FSContainer.stop_container(self.volume_created)
            self.volume_created = None

        if canceled:
            self.set_run_status(RunStatus.CANCELED)



"""
Base class for Whole Tale publishing providers.
"""


class NullManager:
    def updateProgress(self, *args, **kwargs):
        pass


class PublishProvider(object):
    _published = None
    _published_info_index = None

    def __init__(self, gc, tale_id, token, version_id, draft=False, job_manager=None):
        """
        Initialize PublishProvider

        :param gc:  Authenticated Girder client
        :param tale_id:  Tale identifier
        :param token: External Account Token
        :param version_id: The Tale's version that's being published
        :param job_manager:  Optional job manager
        """
        self.gc = gc
        self.draft = draft
        self.token = token
        self.version_id = version_id
        if job_manager is not None:
            self.job_manager = job_manager
        else:
            self.job_manager = NullManager()

        self.tale = self.gc.get("/tale/{}".format(tale_id))
        assert self.tale["description"], "Cannot publish a Tale without a description."
        self.manifest = self.gc.get(
            "/tale/{}/manifest".format(tale_id),
            parameters={"expandFolders": True, "versionId": version_id}
        )

    @property
    def published(self):
        if self._published is not None:
            return self._published

        for i, publish_info in enumerate(self.tale.get("publishInfo", [])):
            if self.resource_server == publish_info.get("repository"):
                self._published = True
                self._published_info_index = i
                break
        else:
            self._published = False
        return self._published

    @property
    def publication_info(self):
        if self.published:
            return self.tale["publishInfo"][self._published_info_index]

    @property
    def access_token(self):
        return self.token["access_token"]

    @property
    def resource_server(self):
        return self.token["resource_server"]

    def publish(self):
        """
        Publish the specified tale using the provided authenticated
        Girder client. If provided, use job_manager to report progress.
        """
        raise NotImplementedError

    def rollback(self, *args, **kwargs):
        """
        Rollback publication process if possible.
        """
        pass



import hashlib
import logging
import os
import tempfile

import docker

from ..utils import (
    DEPLOYMENT,
    DummyTask,
    stop_container,
)
from .builder import ImageBuilderBase


class DockerHelper:
    def __init__(
        self, registry_user=None, registry_password=None, registry_url=None, auth=True
    ):
        username = registry_user or os.environ.get("REGISTRY_USER", "fido")
        password = registry_password or os.environ.get("REGISTRY_PASS")
        registry_url = registry_url or DEPLOYMENT.registry_url
        self.cli = docker.from_env(version="1.28")
        self.apicli = docker.APIClient(base_url="unix://var/run/docker.sock")
        if auth:
            self.cli.login(username=username, password=password, registry=registry_url)
            self.apicli.login(
                username=username, password=password, registry=registry_url
            )


class DockerImageBuilder(ImageBuilderBase):
    def __init__(
        self,
        gc,
        imageId=None,
        tale=None,
        registry_user=None,
        registry_password=None,
        registry_url=None,
        auth=True,
    ):
        super().__init__(gc, imageId=imageId, tale=tale, auth=auth)
        self.dh = DockerHelper(
            registry_user=registry_user,
            registry_password=registry_password,
            registry_url=registry_url,
            auth=auth,
        )

    def pull_r2d(self):
        try:
            self.dh.cli.images.pull(self.container_config.repo2docker_version)
        except docker.errors.NotFound:
            raise ValueError(
                f"Requested r2d image '{self.container_config.repo2docker_version}' not found."
            )

    def push_image(self, image):
        """Push image to the registry"""
        repository, tag = image.split(":", 1)
        for line in self.dh.apicli.push(repository, tag=tag, stream=True, decode=True):
            print(line)

    def run_r2d(self, tag, dry_run=False, task=None):
        """
        Run repo2docker on the workspace using a shared temp directory. Note that
        this uses the "local" provider.  Use the same default user-id and
        user-name as BinderHub
        """

        task = task or DummyTask

        # Extra arguments for r2d
        r2d_cmd = self.r2d_command(tag, dry_run=dry_run)
        r2d_context_dir = os.path.relpath(self.build_context, tempfile.gettempdir())
        host_r2d_context_dir = os.path.join(DEPLOYMENT.tmpdir_mount, r2d_context_dir)

        logging.info("Calling %s", r2d_cmd)

        volumes = {
            "/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"},
            host_r2d_context_dir: {"bind": self.build_context, "mode": "ro"},
        }

        print(f"Using repo2docker {self.container_config.repo2docker_version}")
        container = self.dh.cli.containers.run(
            image=self.container_config.repo2docker_version,
            command=r2d_cmd,
            environment=["DOCKER_HOST=unix:///var/run/docker.sock"],
            privileged=True,
            detach=True,
            remove=False,
            volumes=volumes,
        )

        # Job output must come from stdout/stderr
        h = hashlib.md5("R2D output".encode())
        for line in container.logs(stream=True):
            if task.canceled:
                task.request.chain = None
                stop_container(container)
                break
            output = line.decode("utf-8").strip()
            if not output.startswith("Using local repo"):  # contains variable path
                h.update(output.encode("utf-8"))
            if not dry_run:  # We don't want to see it.
                print(output)

        try:
            ret = container.wait()
        except docker.errors.NotFound:
            ret = {"StatusCode": -123}
        container.remove()

        if ret["StatusCode"] != 0:
            logging.error("Error building image")
        # Since detach=True, then we need to explicitly check for the
        # container exit code
        return ret, h.hexdigest()
